from .download import download_zip, download_file
from pathlib import Path
down_path = Path(__file__).parent
xz_url = "addons_file"+"/"+down_path.name+"/"
download_file(xz_url + "AgedFX.blend",down_path)

import os
import bpy
import colorsys
from bpy.props import StringProperty, FloatProperty, PointerProperty, BoolProperty, FloatVectorProperty, IntProperty
from bpy.types import AddonPreferences, PropertyGroup, Panel, Operator
from mathutils import Vector, Matrix
from .translation.translate import GetTranslationDict


def update_fog_color(self, context):
    node_name = "Fog_AgedFX"
    for material in bpy.data.materials:
        if material.use_nodes:
            node = material.node_tree.nodes.get(node_name)
            if node and "Color" in node.inputs:
                node.inputs["Color"].default_value = self.fog_color
                
def update_sheen_tint_color(self, context):
    for obj in context.selected_objects:
        if obj.type == 'MESH' and obj.material_slots:
            for slot in obj.material_slots:
                material = slot.material
                if material and material.use_nodes:
                    for node in material.node_tree.nodes:
                        if node.type == 'BSDF_PRINCIPLED':
                            if any(group_node for group_node in material.node_tree.nodes if group_node.type == 'GROUP' and group_node.node_tree.name == 'Dust'):
                                node.inputs['Sheen Tint'].default_value = [self.sheen_tint_color[0], self.sheen_tint_color[1], self.sheen_tint_color[2], 1.0]
                                
def update_dirt_colors(self, context):
    selected_objects = context.selected_objects
    for obj in selected_objects:
        if obj.type == 'MESH' and obj.material_slots:
            for slot in obj.material_slots:
                material = slot.material
                if material and material.use_nodes:
                    color_ramp_node = material.node_tree.nodes.get('Dirt_Color_Ramp')
                    if color_ramp_node:
                        color_ramp_node.color_ramp.elements[0].color = self.dirt_color_stop_0
                        color_ramp_node.color_ramp.elements[1].color = self.dirt_color_stop_1
                        
def update_light_ray_color(self, context):
    for obj in bpy.data.objects:
        if obj.type == 'MESH' and obj.material_slots:
            for slot in obj.material_slots:
                material = slot.material
                if material and material.use_nodes:
                    color_ramp_node = material.node_tree.nodes.get('Light_Ray_Color_Ramp')
                    if color_ramp_node:
                        color_ramp_node.color_ramp.elements[1].color = self.light_ray_color_stop_1

class AgedFXPreferences(AddonPreferences):
    bl_idname = __package__
    df_path = Path(__file__).parent/"AgedFX.blend"

    filepath: StringProperty(
        default=str(df_path),
        name="Blend File Path",
        subtype='FILE_PATH',
    )

    def draw(self, context):
        layout = self.layout
        layout.prop(self, "filepath")

class AgedFXProperties(PropertyGroup):

    dust_set: BoolProperty(
        name="Dust Set",
        default=False
    )
    
    dirt_set: BoolProperty(
        name="Dirt Set",
        default=False
    )
    
    metal_set: BoolProperty(
        name="Metal Set",
        default=False
    )
    
    deformations_set: BoolProperty(
        name="Deformations Set",
        default=False
    )
    
    panel_dust_set: BoolProperty(
        name="Panel Dust Set",
        default=False
    )
    
    panel_dirt_set: BoolProperty(
        name="Panel Dirt Set",
        default=False
    )
    
    panel_metal_set: BoolProperty(
        name="Panel Metal Set",
        default=False
    )
    
    panel_deformations_set: BoolProperty(
        name="Panel Deformations Set",
        default=False
    )

    panel_cobweb_set: BoolProperty(
        name="Panel Cobweb Set",
        default=False
    )

    panel_moss_set: BoolProperty(
        name="Panel Moss Set",
        default=False
    )

    panel_fog_set: BoolProperty(
        name="Panel Fog Set",
        default=False
    )

    panel_dust_particles_set: BoolProperty(
        name="Panel Dust Particles Set",
        default=False
    )
    
    panel_moths_set: BoolProperty(
        name="Expand Moths Panel",
        description="Expand or collapse the Moths panel",
        default=False
    )
    
    panel_ants_set: BoolProperty(
        name="Expand Ants Panel",
        description="Expand or collapse the Ants panel",
        default=False
    )
    
    panel_spiders_set: BoolProperty(
        name="Expand Spiders Panel",
        description="Expand or collapse the Spiders panel",
        default=False
    )
    
    panel_light_rays_set: BoolProperty(
        name="Expand Light Rays Panel",
        description="Expand or collapse the Light Rays panel",
        default=False
    )
    
    panel_remove_drivers_set: BoolProperty(
        name="Expand Convert Drivers Panel",
        description="Expand or collapse the Convert Drivers panel",
        default=False
    )
    
    panel_quick_fix_set: BoolProperty(
        name="Expand Quick Fix Panel",
        description="Expand or collapse the Quick Fix panel",
        default=False
    )

    panel_replace_glass_set: BoolProperty(
        name="Expand Replace Glass Panel",
        description="Expand or collapse the Replace Glass panel",
        default=False
    )
    
    hide_volume: BoolProperty(
        name="Hide Volume",
        description="Toggle to hide/show volume objects",
        default=False,
        update=lambda self, context: toggle_volume_visibility(self, context)
    )

    dust_amount: FloatProperty(
        name="Dust Amount",
        default=0.0,
        min=0.0,
        max=1.0
    )

    dust_scale: FloatProperty(
        name="Dust Scale",
        default=0.0,
        min=0.0,
        max=1000.0
    )
    
    dust_top: FloatProperty(
        name="Dust Top Ratio",
        default=0.0,
        min=0.0,
        max=1.0
    )
    
    dust_side: FloatProperty(
        name="Dust Viewing Angle Unconsistency",
        default=0.0,
        min=0.0,
        max=1.0
    )
    
    sheen_tint_color: FloatVectorProperty(
        name="Sheen Tint Color",
        subtype='COLOR',
        default=(0.5, 0.5, 0.5, 1),
        size=4,
        min=0.0,
        max=1.0,
        description="Set the Sheen Tint color for the Principled BSDF",
        update=update_sheen_tint_color
    )

    deformations_amount: FloatProperty(
        name="Deformations Amount",
        default=0.0,
        min=0.0,
        max=1.0
    )

    deformations_scale: FloatProperty(
        name="Deformations Scale",
        default=0.0,
        min=0.0,
        max=1000.0
    )

    scratches: FloatProperty(
        name="Scratches",
        default=0.0,
        min=0.0,
        max=1.0
    )

    bumps: FloatProperty(
        name="Bumps",
        default=0.0,
        min=0.0,
        max=1.0
    )

    general_dirt_scale: FloatProperty(
        name="General Dirt Scale",
        default=0.0,
        min=0.0,
        max=1000.0
    )
    
    edge_dirt_scale: FloatProperty(
        name="Edge Dirt Scale",
        default=0.0,
        min=0.0,
        max=1000.0
    )

    general_dirt: FloatProperty(
        name="General Dirt",
        default=0.0,
        min=0.0,
        max=10.0
    )

    edge_dirt: FloatProperty(
        name="Edge Dirt",
        default=0.0,
        min=0.0,
        max=10.0
    )

    edge_dirt_spread: FloatProperty(
        name="Edge Dirt Spread",
        default=0.0,
        min=0.0,
        max=1.0
    )
    
    dirt_color_stop_0: FloatVectorProperty(
        name="Change Color 1 of Selected",
        subtype='COLOR',
        default=colorsys.hsv_to_rgb(0.050, 1.000, 0.050) + (1.0,),
        size=4,
        min=0.0,
        max=1.0,
        description="Color for the first stop in the dirt color ramp",
        update=update_dirt_colors
    )
    
    dirt_color_stop_1: FloatVectorProperty(
        name="Change Color 2 of Selected",
        subtype='COLOR',
        default=colorsys.hsv_to_rgb(0.050, 1.000, 0.150) + (1.0,),
        size=4,
        min=0.0,
        max=1.0,
        description="Color for the second stop in the dirt color ramp",
        update=update_dirt_colors
    )

    general_metal_scale: FloatProperty(
        name="General Metal Scale",
        default=0.0,
        min=0.0,
        max=1000.0
    )

    edge_metal_scale: FloatProperty(
        name="Edge Metal Scale",
        default=0.0,
        min=0.0,
        max=1000.0
    )

    general_metal: FloatProperty(
        name="General Metal",
        default=0.0,
        min=0.0,
        max=10.0
    )

    edge_metal: FloatProperty(
        name="Edge Metal",
        default=0.0,
        min=0.0,
        max=10.0
    )

    edge_metal_spread: FloatProperty(
        name="Edge Metal Spread",
        default=0.0,
        min=0.0,
        max=1.0
    )

    cobweb_translucency: FloatProperty(
        name="Translucency",
        default=0.0,
        min=0.0,
        max=1.0
    )

    apply_scale: BoolProperty(
        name="Apply Scale",
        description="Apply scale to selected objects before adding the Moss Generator",
        default=True
    )

    fog_density: FloatProperty(
        name="Density",
        default=0.0,
        min=0.0,
        max=1.0
    )

    fog_anisotropy: FloatProperty(
        name="Anisotropy",
        default=0.0,
        min=-1.0,
        max=1.0
    )
    
    fog_x_scale: FloatProperty(
        name="Scale X",
        default=0.0,
        min=0.0,
        max=1000.0
    )
    
    fog_y_scale: FloatProperty(
        name="Scale Y",
        default=0.0,
        min=0.0,
        max=1000.0
    )
    
    fog_z_scale: FloatProperty(
        name="Scale Z",
        default=0.0,
        min=0.0,
        max=1000.0
    )
    
    ground_fog_density: FloatProperty(
        name="Ground Fog Density",
        default=0.0,
        min=0.0,
        max=100.0
    )
    
    ground_fog_scale: FloatProperty(
        name="Ground Fog Scale",
        default=0.0,
        min=0.0,
        max=1000.0
    )
    
    ground_fog_height: FloatProperty(
        name="Ground Fog Height",
        default=0.0,
        min=0.0,
        max=1000.0
    )
    
    ground_fog_position_x: FloatProperty(
        name="Ground Fog Position X",
        default=0.0,
        min=-1000.0,
        max=1000.0
    )
    
    ground_fog_position_y: FloatProperty(
        name="Ground Fog Position Y",
        default=0.0,
        min=-1000.0,
        max=1000.0
    )
    
    ground_fog_speed_x: FloatProperty(
        name="Ground Fog Speed X",
        default=0.0,
        min=0.0,
        max=1.0
    )
    
    ground_fog_speed_y: FloatProperty(
        name="Ground Fog Speed Y",
        default=0.0,
        min=0.0,
        max=1.0
    )
    
    ground_fog_turbolence: FloatProperty(
        name="Ground Fog Turbolence",
        default=0.0,
        min=0.0,
        max=1.0
    )
    
    fog_color: FloatVectorProperty(
        name="Color",
        subtype='COLOR',
        default=(0.5, 0.5, 0.5, 1),
        size=4,
        min=0.0,
        max=1.0,
        description="Choose a custom color",
        update=update_fog_color
    )
    
    dust_particles_size_x: FloatProperty(
        name="Domain Size X",
        default=0.0,
        min=0.0,
        max=1000.0
    )

    dust_particles_size_y: FloatProperty(
        name="Domain Size Y",
        default=0.0,
        min=0.0,
        max=1000.0
    )

    dust_particles_size_z: FloatProperty(
        name="Domain Size Z",
        default=0.0,
        min=0.0,
        max=1000.0
    )
    
    dust_particles_wind_x: FloatProperty(
        name="Wind X",
        default=0.0,
        min=0.0,
        max=10.0
    )
    
    dust_particles_wind_y: FloatProperty(
        name="Wind Y",
        default=0.0,
        min=0.0,
        max=10.0
    )
    
    dust_particles_wind_z: FloatProperty(
        name="Wind Z",
        default=0.0,
        min=0.0,
        max=10.0
    )
    
    dust_particles_amount: FloatProperty(
        name="Particles Amount",
        default=0.0,
        min=0.0,
        max=1000.0
    )
    
    dust_particles_scale_min: FloatProperty(
        name="Particles Scale Min",
        default=0.0,
        min=0.0,
        max=10.0
    )
    
    dust_particles_scale_max: FloatProperty(
        name="Particles Scale Max",
        default=0.0,
        min=0.0,
        max=10.0
    )
    
    dust_particles_speed: FloatProperty(
        name="Particles Speed",
        default=0.0,
        min=0.0,
        max=50.0
    )
    
    dust_particles_rotation: FloatProperty(
        name="Particles Rotation",
        default=0.0,
        min=0.0,
        max=1.0
    )
    
    dust_particles_turbolence: FloatProperty(
        name="Particles Turbolence",
        default=0.0,
        min=0.0,
        max=15.0
    )
    
    dust_particles_seed: FloatProperty(
        name="Particles Seed",
        default=0,
        min=0,
        max=100,
        step=100,
    )
    
    dust_particles_emission: FloatProperty(
        name="Particles Emission",
        default=0.0,
        min=0.0,
        max=1.0
    )
        
    light_ray_light: FloatProperty(
        name="Light Strength",
        default=80.0,
        min=0.0,
        max=1000.0
    )
    
    light_ray_volume: FloatProperty(
        name="Volume Strength",
        default=0.2,
        min=0.0,
        max=10.0
    )
    
    light_ray_color_stop_1: FloatVectorProperty(
        name="Color",
        subtype='COLOR',
        default=colorsys.hsv_to_rgb(0.000, 0.000, 1.000) + (1.0,),
        size=4,
        min=0.0,
        max=1.0,
        description="Change the color of the light rays.",
        update=update_light_ray_color
    )

class AgedFXPanel(Panel):
    bl_label = "AgedFX"
    bl_idname = "AGEDFX_PT_main_panel"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = "AgedFX"

    def draw(self, context):
        layout = self.layout
        row = layout.row()
        row.label(text="v 1.5")
        row.operator("wm.url_open", text="", icon='HELP').url = "https://cgb3d.taobao.com"
        
        scene = context.scene
        aged_fx = scene.aged_fx

        quick_fix_box = layout.box()
        row = quick_fix_box.row()
        row.prop(aged_fx, "panel_quick_fix_set", text="", icon="TRIA_DOWN" if aged_fx.panel_quick_fix_set else "TRIA_RIGHT", emboss=False)
        row.label(text="Quick Fix")

        if aged_fx.panel_quick_fix_set:
            quick_fix_box.operator("object.apply_scale", text="Apply Scale")
            quick_fix_box.operator("object.make_single_user_data", text="Make Single User Data")

        glass_box = layout.box()
        row = glass_box.row()
        row.prop(aged_fx, "panel_replace_glass_set", text="", icon="TRIA_DOWN" if aged_fx.panel_replace_glass_set else "TRIA_RIGHT", emboss=False)
        row.label(text="Replace Glass")

        if aged_fx.panel_replace_glass_set:
            glass_box.operator("object.replace_glass_all", text="All")
            glass_box.operator("object.replace_glass_selected", text="Selected")
        
        remove_drivers_box = layout.box()
        row = remove_drivers_box.row()
        row.prop(aged_fx, "panel_remove_drivers_set", text="", icon="TRIA_DOWN" if aged_fx.panel_remove_drivers_set else "TRIA_RIGHT", emboss=False)
        row.label(text="Convert Drivers")

        if aged_fx.panel_remove_drivers_set:
            row = remove_drivers_box.row()
            row.operator("object.remove_all_drivers", text="Convert to Static")
        
        layout.separator()

        dust_group_users = self.count_dust_group_users()
        dust_icon = 'SHADING_SOLID' if self.group_exists_in_material("Dust") else 'ANTIALIASED'
        dust_box = layout.box()
        row = dust_box.row()
        row.prop(aged_fx, "panel_dust_set", text="", icon="TRIA_DOWN" if aged_fx.panel_dust_set else "TRIA_RIGHT", emboss=False)
        row.label(text=f"表面灰尘 ({dust_group_users})", icon=dust_icon)

        if aged_fx.panel_dust_set:
            row = dust_box.row()
            row.operator("object.make_dust_operator")
            row.operator("object.delete_dust_operator")
            row = dust_box.row()
            row.operator("object.make_dust_selected_operator")
            row.operator("object.delete_dust_selected_operator")
            row = dust_box.row()
            row.operator("object.make_unique_dust")
            dust_box.prop(aged_fx, "dust_amount")
            dust_box.prop(aged_fx, "dust_scale")
            dust_box.prop(aged_fx, "dust_top")
            dust_box.prop(aged_fx, "dust_side")
            dust_box.prop(aged_fx, "sheen_tint_color", text="Change Color of Selected")

        dirt_group_users = self.count_dirt_group_users()
        dirt_icon = 'SHADING_SOLID' if self.group_exists_in_material("Dirt") else 'ANTIALIASED'
        dirt_box = layout.box()
        row = dirt_box.row()
        row.prop(aged_fx, "panel_dirt_set", text="", icon="TRIA_DOWN" if aged_fx.panel_dirt_set else "TRIA_RIGHT", emboss=False)
        row.label(text=f"污垢 ({dirt_group_users})", icon=dirt_icon)

        if aged_fx.panel_dirt_set:
            row = dirt_box.row()
            row.operator("object.make_dirt_operator")
            row.operator("object.delete_dirt_operator")
            row = dirt_box.row()
            row.operator("object.make_dirt_selected_operator")
            row.operator("object.delete_dirt_selected_operator")
            row = dirt_box.row()
            row.operator("object.make_unique_dirt")
            dirt_box.prop(aged_fx, "general_dirt")
            dirt_box.prop(aged_fx, "general_dirt_scale")
            dirt_box.prop(aged_fx, "edge_dirt")
            dirt_box.prop(aged_fx, "edge_dirt_scale")
            dirt_box.prop(aged_fx, "edge_dirt_spread")
            dirt_box.prop(aged_fx, "dirt_color_stop_0", text="Change Color 1 of Selected")
            dirt_box.prop(aged_fx, "dirt_color_stop_1", text="Change Color 2 of Selected")
            
        metal_group_users = self.count_metal_group_users()
        metal_icon = 'SHADING_SOLID' if self.group_exists_in_material("Metal") else 'ANTIALIASED'
        metal_box = layout.box()
        row = metal_box.row()
        row.prop(aged_fx, "panel_metal_set", text="", icon="TRIA_DOWN" if aged_fx.panel_metal_set else "TRIA_RIGHT", emboss=False)
        row.label(text=f"金属 ({metal_group_users})", icon=metal_icon)

        if aged_fx.panel_metal_set:
            row = metal_box.row()
            row.operator("object.make_metal_operator")
            row.operator("object.delete_metal_operator")
            row = metal_box.row()
            row.operator("object.make_metal_selected_operator")
            row.operator("object.delete_metal_selected_operator")
            row = metal_box.row()
            row.operator("object.make_unique_metal")
            metal_box.prop(aged_fx, "general_metal")
            metal_box.prop(aged_fx, "general_metal_scale")
            metal_box.prop(aged_fx, "edge_metal")
            metal_box.prop(aged_fx, "edge_metal_scale")
            metal_box.prop(aged_fx, "edge_metal_spread")
            metal_box.separator()
            row = metal_box.row()
            row.operator("object.make_selected_rusty", text="Make Selected Rusty")
            row = metal_box.row()
            row.operator("object.undo_rusty", text="Undo")

        deformations_group_users = self.count_deformations_group_users()
        deformations_icon = 'SHADING_SOLID' if self.group_exists_in_material("Deformations") else 'ANTIALIASED'
        deformations_box = layout.box()
        row = deformations_box.row()
        row.prop(aged_fx, "panel_deformations_set", text="", icon="TRIA_DOWN" if aged_fx.panel_deformations_set else "TRIA_RIGHT", emboss=False)
        row.label(text=f"痕迹 ({deformations_group_users})", icon=deformations_icon)

        if aged_fx.panel_deformations_set:
            row = deformations_box.row()
            row.operator("object.make_deformations_operator")
            row.operator("object.delete_deformations_operator")
            row = deformations_box.row()
            row.operator("object.make_deformations_selected_operator")
            row.operator("object.delete_deformations_selected_operator")
            row = deformations_box.row()
            row.operator("object.make_unique_deformations")
            deformations_box.prop(aged_fx, "deformations_amount")
            deformations_box.prop(aged_fx, "deformations_scale")
            deformations_box.prop(aged_fx, "scratches")
            deformations_box.prop(aged_fx, "bumps")
            
        layout.separator()
        
        fog_box = layout.box()
        row = fog_box.row()
        row.prop(aged_fx, "panel_fog_set", text="", icon="TRIA_DOWN" if aged_fx.panel_fog_set else "TRIA_RIGHT", emboss=False)
        row.label(text="Fog")

        if aged_fx.panel_fog_set:
            row = fog_box.row()
            row.operator("object.add_fog", text="Add")
            row.operator("object.delete_fog", text="Delete")
            fog_box.prop(aged_fx, "fog_density")
            fog_box.prop(aged_fx, "fog_anisotropy")
            fog_box.prop(aged_fx, "fog_x_scale")
            fog_box.prop(aged_fx, "fog_y_scale")
            fog_box.prop(aged_fx, "fog_z_scale")
            fog_box.prop(aged_fx, "ground_fog_density")
            fog_box.prop(aged_fx, "ground_fog_scale")
            fog_box.prop(aged_fx, "ground_fog_height")
            fog_box.prop(aged_fx, "ground_fog_position_x")
            fog_box.prop(aged_fx, "ground_fog_position_y")
            fog_box.prop(aged_fx, "ground_fog_speed_x")
            fog_box.prop(aged_fx, "ground_fog_speed_y")
            fog_box.prop(aged_fx, "ground_fog_turbolence")
            fog_box.prop(aged_fx, "fog_color", text="Color")
        
        light_rays_box = layout.box()
        row = light_rays_box.row()
        row.prop(aged_fx, "panel_light_rays_set", text="", icon="TRIA_DOWN" if aged_fx.panel_light_rays_set else "TRIA_RIGHT", emboss=False)
        row.label(text="Light Rays")

        if aged_fx.panel_light_rays_set:
            row = light_rays_box.row()
            row.operator("object.add_constant_light_ray", text="Add Constant")
            row = light_rays_box.row()
            row.operator("object.add_dissolving_light_ray", text="Add Dissolving")
            row = light_rays_box.row()
            row.operator("object.delete_light_ray", text="Delete")
            light_rays_box.separator()
            light_rays_box.prop(aged_fx, "light_ray_light")
            light_rays_box.prop(aged_fx, "light_ray_volume")
            light_rays_box.prop(aged_fx, "hide_volume", text="Hide Volume", toggle=True)
            light_rays_box.prop(aged_fx, "light_ray_color_stop_1", text="Color")
            
        dust_particles_box = layout.box()
        row = dust_particles_box.row()
        row.prop(aged_fx, "panel_dust_particles_set", text="", icon="TRIA_DOWN" if aged_fx.panel_dust_particles_set else "TRIA_RIGHT", emboss=False)
        row.label(text="Dust Particles")

        if aged_fx.panel_dust_particles_set:
            row = dust_particles_box.row()
            row.operator("object.add_dust_particles", text="Add")
            row.operator("object.delete_dust_particles", text="Delete")
            dust_particles_box.prop(aged_fx, "dust_particles_size_x")
            dust_particles_box.prop(aged_fx, "dust_particles_size_y")
            dust_particles_box.prop(aged_fx, "dust_particles_size_z")
            dust_particles_box.prop(aged_fx, "dust_particles_wind_x")
            dust_particles_box.prop(aged_fx, "dust_particles_wind_y")
            dust_particles_box.prop(aged_fx, "dust_particles_wind_z")
            dust_particles_box.prop(aged_fx, "dust_particles_amount")
            dust_particles_box.prop(aged_fx, "dust_particles_scale_min")
            dust_particles_box.prop(aged_fx, "dust_particles_scale_max")
            dust_particles_box.prop(aged_fx, "dust_particles_speed")
            dust_particles_box.prop(aged_fx, "dust_particles_rotation")
            dust_particles_box.prop(aged_fx, "dust_particles_turbolence")
            dust_particles_box.prop(aged_fx, "dust_particles_seed")
            dust_particles_box.prop(aged_fx, "dust_particles_emission")
            
        layout.separator()

        cobweb_box = layout.box()
        row = cobweb_box.row()
        row.prop(aged_fx, "panel_cobweb_set", text="", icon="TRIA_DOWN" if aged_fx.panel_cobweb_set else "TRIA_RIGHT", emboss=False)
        row.label(text="Cobwebs")

        if aged_fx.panel_cobweb_set:
            row = cobweb_box.row()
            row.operator("object.add_cobweb", text="Add")
            row.operator("object.delete_cobweb", text="Delete")
            cobweb_box.prop(aged_fx, "cobweb_translucency")
            row = cobweb_box.row()
            row.operator("object.convert_selected_to_mesh", text="Convert Selected to Mesh")

        moss_box = layout.box()
        row = moss_box.row()
        row.prop(aged_fx, "panel_moss_set", text="", icon="TRIA_DOWN" if aged_fx.panel_moss_set else "TRIA_RIGHT", emboss=False)
        row.label(text="Moss")

        if aged_fx.panel_moss_set:
            row = moss_box.row()
            row.prop(aged_fx, "apply_scale", text="Apply Scale", toggle=True)
            row = moss_box.row()
            row.operator("object.add_moss_generator", text="Add Selected")
            row.operator("object.remove_moss_generator", text="Delete Selected")
            
        layout.separator()
            
        moths_box = layout.box()
        row = moths_box.row()
        row.prop(aged_fx, "panel_moths_set", text="", icon="TRIA_DOWN" if aged_fx.panel_moths_set else "TRIA_RIGHT", emboss=False)
        row.label(text="Moths")

        if aged_fx.panel_moths_set:
            row = moths_box.row()
            row.operator("object.add_moths", text="Add")
            row.operator("object.delete_moths", text="Delete")
            row = moths_box.row()
            row.operator("object.add_collision_modifier", text="Add Collision Modifier")
            row = moths_box.row()
            row.operator("object.remove_collision_modifier", text="Remove Collision Modifier")
            
        ants_box = layout.box()
        row = ants_box.row()
        row.prop(aged_fx, "panel_ants_set", text="", icon="TRIA_DOWN" if aged_fx.panel_ants_set else "TRIA_RIGHT", emboss=False)
        row.label(text="Ants")

        if aged_fx.panel_ants_set:
            row = ants_box.row()
            row.operator("object.add_ants", text="Add")
            row.operator("object.delete_ants", text="Delete")
            row = ants_box.row()
            row.operator("object.apply_transforms_selected", text="Apply Transforms")
            row = ants_box.row()
            row.operator("object.add_shrinkwrap_plane_ants", text="Add Shrinkwrap Plane")
            
        spiders_box = layout.box()
        row = spiders_box.row()
        row.prop(aged_fx, "panel_spiders_set", text="", icon="TRIA_DOWN" if aged_fx.panel_spiders_set else "TRIA_RIGHT", emboss=False)
        row.label(text="Spiders")

        if aged_fx.panel_spiders_set:
            row = spiders_box.row()
            row.operator("object.add_spiders", text="Add")
            row.operator("object.delete_spiders", text="Delete")
            row = spiders_box.row()
            row.operator("object.apply_transforms_selected", text="Apply Transforms")
            row = spiders_box.row()
            row.operator("object.add_shrinkwrap_plane_spiders", text="Add Shrinkwrap Plane")
        
    def group_exists_in_material(self, group_name):
        for mat in bpy.data.materials:
            if mat.node_tree:
                for node in mat.node_tree.nodes:
                    if node.type == 'GROUP' and node.node_tree and node.node_tree.name == group_name:
                        return True
        return False

    def count_dust_group_users(self):
        return self.count_group_users("Dust")

    def count_dirt_group_users(self):
        return self.count_group_users("Dirt")
        
    def count_metal_group_users(self):
        return self.count_group_users("Metal")

    def count_deformations_group_users(self):
        return self.count_group_users("Deformations")

    def count_group_users(self, group_name):
        count = 0
        for mat in bpy.data.materials:
            if mat.node_tree:
                for node in mat.node_tree.nodes:
                    if node.type == 'GROUP' and node.node_tree and node.node_tree.name == group_name:
                        count += 1
        return count
        
class OBJECT_OT_toggle_apply_scale(bpy.types.Operator):
    """Toggle the apply scale option"""
    bl_idname = "agedfx.toggle_apply_scale"
    bl_label = "Toggle Apply Scale"

    def execute(self, context):
        context.scene.aged_fx.apply_scale = not context.scene.aged_fx.apply_scale
        return {'FINISHED'}
        
class OBJECT_OT_ApplyScale(Operator):
    bl_idname = "object.apply_scale"
    bl_label = "Apply Scale"
    bl_description = "Apply scale to all objects with attached Dust, Dirt, Metal or Deformations"

    def execute(self, context):
        node_groups = ["Dust", "Dirt", "Metal", "Deformations"]
        for obj in bpy.data.objects:
            if obj.type == 'MESH' and obj.visible_get():
                for slot in obj.material_slots:
                    if slot.material and slot.material.use_nodes:
                        if any(node for node in slot.material.node_tree.nodes if node.bl_idname == 'ShaderNodeGroup' and node.node_tree and node.node_tree.name in node_groups):
                            bpy.context.view_layer.objects.active = obj
                            bpy.ops.object.select_all(action='DESELECT')
                            obj.select_set(True)
                            bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)
                            obj.select_set(False)
        return {'FINISHED'}
        
class OBJECT_OT_MakeSingleUserData(Operator):
    bl_idname = "object.make_single_user_data"
    bl_label = "Make Single User Data"
    bl_description = "Make linked object data local to all objects with attached Dust, Dirt, Metal or Deformations"

    def execute(self, context):
        node_groups = ["Dust", "Dirt", "Metal", "Deformations"]
        for obj in bpy.data.objects:
            if obj.type == 'MESH' and obj.visible_get():
                for slot in obj.material_slots:
                    if slot.material and slot.material.use_nodes:
                        if any(node for node in slot.material.node_tree.nodes if node.bl_idname == 'ShaderNodeGroup' and node.node_tree and node.node_tree.name in node_groups):
                            if obj.data.users > 1:
                                obj.data = obj.data.copy()
        return {'FINISHED'}
        
class OBJECT_OT_ReplaceGlassAll(Operator):
    bl_idname = "object.replace_glass_all"
    bl_label = "All"
    bl_description = "Replace Glass BSDFs with Principled BSDFs in all materials"

    def execute(self, context):
        replace_glass_bsdf_in_materials(bpy.data.materials)
        return {'FINISHED'}

class OBJECT_OT_ReplaceGlassSelected(Operator):
    bl_idname = "object.replace_glass_selected"
    bl_label = "Selected"
    bl_description = "Replace Glass BSDFs with Principled BSDFs in selected objects"

    def execute(self, context):
        selected_materials = get_materials_from_selected_objects()
        replace_glass_bsdf_in_materials(selected_materials)
        return {'FINISHED'}

class OBJECT_OT_MakeDustButton(Operator):
    bl_idname = "object.make_dust_operator"
    bl_label = "Add All"
    bl_description = "Click this button to add dust"

    def execute(self, context):
        if not context.scene.aged_fx.dust_set:
            context.scene.aged_fx.dust_amount = 0.5
            context.scene.aged_fx.dust_scale = 1.0
            context.scene.aged_fx.dust_top = 0.5
            context.scene.aged_fx.dust_side = 0.5
            context.scene.aged_fx.sheen_tint_color = (0.5, 0.5, 0.5, 1.0)
            context.scene.aged_fx.dust_set = True
        make_dust(self, context)
        
        objects = bpy.data.objects
        for obj in objects:
            for slot in obj.material_slots:
                if slot.material and slot.material.use_nodes:
                    add_dust_attributes_to_material(slot.material)
                    
        return {'FINISHED'}
        
class OBJECT_OT_MakeDustSelectedButton(Operator):
    bl_idname = "object.make_dust_selected_operator"
    bl_label = "Add Selected"
    bl_description = "Click this button to add dust to selected objects"

    def execute(self, context):
        if not context.scene.aged_fx.dust_set:
            context.scene.aged_fx.dust_amount = 0.5
            context.scene.aged_fx.dust_scale = 1.0
            context.scene.aged_fx.dust_top = 0.5
            context.scene.aged_fx.dust_side = 0.5
            context.scene.aged_fx.sheen_tint_color = (0.5, 0.5, 0.5, 1.0)
            context.scene.aged_fx.dust_set = True
        make_dust_selected(self, context)
        
        selected_objects = context.selected_objects
        for obj in selected_objects:
            for slot in obj.material_slots:
                if slot.material and slot.material.use_nodes:
                    add_dust_attributes_to_material(slot.material)
                    
        return {'FINISHED'}
        
class OBJECT_OT_DeleteDustButton(Operator):
    bl_idname = "object.delete_dust_operator"
    bl_label = "Delete All"
    bl_description = "Click this button to delete dust"

    def execute(self, context):
        objects = set(obj for obj in bpy.data.objects)
        materials_done = set()

        my_node_group = bpy.data.node_groups.get('Dust')

        if not my_node_group:
            return {'FINISHED'}

        for obj in objects:
            if obj.type != 'MESH':
                continue

            for idx, slot in enumerate(obj.material_slots):
                if slot.material is None:
                    continue
                if slot.material.name in materials_done:
                    continue
                if not slot.material.use_nodes:
                    continue

                if material_uses_group(slot.material, my_node_group):
                    if remove_node_group_from_material(slot.material, my_node_group):
                        materials_done.add(slot.material.name)

                nodes = slot.material.node_tree.nodes
                for node in nodes:
                    if node.label == "Dust Amount":
                        nodes.remove(node)
                        break

        context.scene.aged_fx.dust_amount = 0.0
        context.scene.aged_fx.dust_scale = 0.0
        context.scene.aged_fx.dust_top = 0.0
        context.scene.aged_fx.dust_side = 0.0
        context.scene.aged_fx.sheen_tint_color = (0.5, 0.5, 0.5, 1.0)
        context.scene.aged_fx.dust_set = False
        
        objects = bpy.data.objects
        for obj in objects:
            for slot in obj.material_slots:
                if slot.material and slot.material.use_nodes:
                    remove_dust_attributes_from_material(slot.material)
        
        return {'FINISHED'}
        
class OBJECT_OT_DeleteDustSelectedButton(Operator):
    bl_idname = "object.delete_dust_selected_operator"
    bl_label = "Delete Selected"
    bl_description = "Click this button to delete dust from selected objects"

    def execute(self, context):
        selected_objects = [obj for obj in context.selected_objects if obj.type == 'MESH']
        materials_done = set()

        my_node_group = bpy.data.node_groups.get('Dust')
        if not my_node_group:
            return {'FINISHED'}

        for obj in selected_objects:
            for slot in obj.material_slots:
                if slot.material is None or slot.material.name in materials_done or not slot.material.use_nodes:
                    continue

                nodes = slot.material.node_tree.nodes
                for node in nodes:
                    if node.label == "Dust Amount":
                        nodes.remove(node)
                        break

                if material_uses_group(slot.material, my_node_group):
                    if remove_node_group_from_material(slot.material, my_node_group):
                        materials_done.add(slot.material.name)
                        
        selected_objects = context.selected_objects
        for obj in selected_objects:
            for slot in obj.material_slots:
                if slot.material and slot.material.use_nodes:
                    remove_dust_attributes_from_material(slot.material)
                                                                
        return {'FINISHED'}
        
class OBJECT_OT_MakeUniqueDust(bpy.types.Operator):
    bl_idname = "object.make_unique_dust"
    bl_label = "Make Unique"
    bl_description = "Remove drivers for dust properties on selected objects"

    def execute(self, context):
        for obj in context.selected_objects:
            if obj.type == 'MESH':
                for slot in obj.material_slots:
                    material = slot.material
                    if material and material.use_nodes:
                        for node in material.node_tree.nodes:
                            if node.type == 'VALUE' and node.label in ["Dust Amount", "Dust Scale", "Dust Top Ratio", "Dust Viewing Angle Unconsistency"]:
                                data_path = f'nodes["{node.name}"].outputs[0].default_value'
                                if material.node_tree.animation_data and material.node_tree.animation_data.drivers:
                                    for driver in material.node_tree.animation_data.drivers:
                                        if driver.data_path == data_path:
                                            material.node_tree.driver_remove(driver.data_path)
                                            break

        return {'FINISHED'}
        
class OBJECT_OT_MakeDirtButton(bpy.types.Operator):
    bl_idname = "object.make_dirt_operator"
    bl_label = "Add All"
    bl_description = "Click this button to add dirt"

    def execute(self, context):
        addon_prefs = context.preferences.addons[__package__].preferences
        filepath = addon_prefs.filepath

        my_node_group = bpy.data.node_groups.get('Dirt', None)
        if not my_node_group:
            with bpy.data.libraries.load(filepath) as (data_from, data_to):
                data_to.node_groups = ['Dirt']
            my_node_group = bpy.data.node_groups.get('Dirt', None)
            
            if not my_node_group:
                self.report({'ERROR'}, "The 'Dirt' node group could not be imported. Please ensure the path is correct in the Addon Preferences.")
                return {'CANCELLED'}

        if not context.scene.aged_fx.dirt_set:
            context.scene.aged_fx.general_dirt_scale = 5.0
            context.scene.aged_fx.edge_dirt_scale = 5.0
            context.scene.aged_fx.general_dirt = 8.0
            context.scene.aged_fx.edge_dirt = 4.0
            context.scene.aged_fx.edge_dirt_spread = 0.25
            context.scene.aged_fx.dirt_color_stop_0 = colorsys.hsv_to_rgb(0.050, 1.000, 0.050) + (1.0,)
            context.scene.aged_fx.dirt_color_stop_1 = colorsys.hsv_to_rgb(0.050, 1.000, 0.150) + (1.0,)
            context.scene.aged_fx.dirt_set = True

        excluded_collections = {"Moths AgedFX", "Ants AgedFX", "Spiders AgedFX"}
        objects = set(obj for obj in bpy.data.objects if obj.type == 'MESH' and not is_object_in_excluded_collections(obj, excluded_collections))
        materials_done = set()

        for obj in objects:
            for slot in obj.material_slots:
                material = slot.material
                if not material or not material.use_nodes:
                    continue
                
                if material.name == "Cobweb Generator":
                    continue

                if any(node for node in material.node_tree.nodes if node.type == 'GROUP' and node.node_tree == my_node_group):
                    continue

                if material.name in materials_done:
                    continue

                nodes = material.node_tree.nodes
                links = material.node_tree.links

                pbsdf = find_original_principled_bsdf(material)
                if pbsdf is None:
                    continue

                material_output = next((node for node in nodes if node.type == 'OUTPUT_MATERIAL'), None)
                if material_output is None:
                    continue

                last_mix_shader = None
                for link in links:
                    if link.to_node == material_output and link.from_node.type == 'MIX_SHADER':
                        last_mix_shader = link.from_node
                        break

                noise_texture = nodes.new(type='ShaderNodeTexNoise')
                noise_texture.location = material_output.location + Vector([1600, -250])
                noise_texture.inputs['Scale'].default_value = 100
                noise_texture.inputs['Detail'].default_value = 15
                noise_texture.inputs['Roughness'].default_value = 1
                noise_texture.inputs['Lacunarity'].default_value = 2
                noise_texture.inputs['Distortion'].default_value = 0
                noise_texture.name = "Dirt_Noise_Texture"

                color_ramp = nodes.new(type='ShaderNodeValToRGB')
                color_ramp.location = material_output.location + Vector([1800, -250])
                color_ramp.color_ramp.elements[0].color = colorsys.hsv_to_rgb(0.050, 1.000, 0.050) + (1.0,)
                color_ramp.color_ramp.elements[1].color = colorsys.hsv_to_rgb(0.050, 1.000, 0.150) + (1.0,)
                color_ramp.color_ramp.elements[0].position = 0.4
                color_ramp.color_ramp.elements[1].position = 0.6
                color_ramp.name = "Dirt_Color_Ramp"

                links.new(noise_texture.outputs['Fac'], color_ramp.inputs['Fac'])

                mix_shader = nodes.new(type='ShaderNodeMixShader')
                mix_shader.location = material_output.location + Vector([2400, -250])
                mix_shader.name = "Dirt_Mix_Shader"

                principled_bsdf_copy = nodes.new(type='ShaderNodeBsdfPrincipled')
                principled_bsdf_copy.location = material_output.location + Vector([2100, -250])
                principled_bsdf_copy.name = "Dirt_Principled_BSDF"
                links.new(color_ramp.outputs['Color'], principled_bsdf_copy.inputs['Base Color'])

                if last_mix_shader:
                    links.new(last_mix_shader.outputs['Shader'], mix_shader.inputs[1])
                else:
                    links.new(pbsdf.outputs['BSDF'], mix_shader.inputs[1])
                links.new(principled_bsdf_copy.outputs['BSDF'], mix_shader.inputs[2])

                dirt_node = nodes.new(type='ShaderNodeGroup')
                dirt_node.node_tree = my_node_group
                dirt_node.location = material_output.location + Vector([1400, -250])
                links.new(dirt_node.outputs['Value'], mix_shader.inputs[0])

                links.new(mix_shader.outputs['Shader'], material_output.inputs['Surface'])

                materials_done.add(material.name)

        for obj in bpy.data.objects:
            for slot in obj.material_slots:
                if slot.material and slot.material.use_nodes:
                    add_dirt_attributes_to_material(slot.material)

        return {'FINISHED'}
        
class OBJECT_OT_MakeDirtSelectedButton(bpy.types.Operator):
    bl_idname = "object.make_dirt_selected_operator"
    bl_label = "Add Selected"
    bl_description = "Click this button to add dirt to selected materials"

    def execute(self, context):
        addon_prefs = context.preferences.addons[__package__].preferences
        filepath = addon_prefs.filepath

        my_node_group = bpy.data.node_groups.get('Dirt', None)
        if not my_node_group:
            with bpy.data.libraries.load(filepath) as (data_from, data_to):
                data_to.node_groups = ['Dirt']
            my_node_group = bpy.data.node_groups.get('Dirt', None)
            
            if not my_node_group:
                self.report({'ERROR'}, "The 'Dirt' node group could not be imported. Please ensure the path is correct in the Addon Preferences.")
                return {'CANCELLED'}

        if not context.scene.aged_fx.dirt_set:
            context.scene.aged_fx.general_dirt_scale = 5.0
            context.scene.aged_fx.edge_dirt_scale = 5.0
            context.scene.aged_fx.general_dirt = 8.0
            context.scene.aged_fx.edge_dirt = 4.0
            context.scene.aged_fx.edge_dirt_spread = 0.25
            context.scene.aged_fx.dirt_color_stop_0 = colorsys.hsv_to_rgb(0.050, 1.000, 0.050) + (1.0,)
            context.scene.aged_fx.dirt_color_stop_1 = colorsys.hsv_to_rgb(0.050, 1.000, 0.150) + (1.0,)
            context.scene.aged_fx.dirt_set = True

        selected_objects = [obj for obj in context.selected_objects if obj.type == 'MESH']
        for obj in selected_objects:
            for mat_slot in obj.material_slots:
                material = mat_slot.material
                if not material or not material.use_nodes:
                    continue
                    
                if material.name == "Cobweb Generator":
                    continue

                if any(node for node in material.node_tree.nodes if node.type == 'GROUP' and node.node_tree == my_node_group):
                    continue

                nodes = material.node_tree.nodes
                links = material.node_tree.links

                pbsdf = find_original_principled_bsdf(material)
                if pbsdf is None:
                    continue

                material_output = next((node for node in nodes if node.type == 'OUTPUT_MATERIAL'), None)
                if material_output is None:
                    continue

                noise_texture = nodes.new(type='ShaderNodeTexNoise')
                noise_texture.location = material_output.location + Vector([1600, -250])
                noise_texture.inputs['Scale'].default_value = 100
                noise_texture.inputs['Detail'].default_value = 15
                noise_texture.inputs['Roughness'].default_value = 1
                noise_texture.inputs['Lacunarity'].default_value = 2
                noise_texture.inputs['Distortion'].default_value = 0
                noise_texture.name = "Dirt_Noise_Texture"

                color_ramp = nodes.new(type='ShaderNodeValToRGB')
                color_ramp.location = material_output.location + Vector([1800, -250])
                color_ramp.color_ramp.elements[0].color = colorsys.hsv_to_rgb(0.050, 1.000, 0.050) + (1.0,)
                color_ramp.color_ramp.elements[1].color = colorsys.hsv_to_rgb(0.050, 1.000, 0.150) + (1.0,)
                color_ramp.color_ramp.elements[0].position = 0.4
                color_ramp.color_ramp.elements[1].position = 0.6
                color_ramp.name = "Dirt_Color_Ramp"

                links.new(noise_texture.outputs['Fac'], color_ramp.inputs['Fac'])

                mix_shader = nodes.new(type='ShaderNodeMixShader')
                mix_shader.location = material_output.location + Vector([2400, -250])
                mix_shader.name = "Dirt_Mix_Shader"

                principled_bsdf_copy = nodes.new(type='ShaderNodeBsdfPrincipled')
                principled_bsdf_copy.location = material_output.location + Vector([2100, -250])
                principled_bsdf_copy.name = "Dirt_Principled_BSDF"
                links.new(color_ramp.outputs['Color'], principled_bsdf_copy.inputs['Base Color'])

                last_mix_shader = None
                for link in links:
                    if link.to_node == material_output and link.from_node.type == 'MIX_SHADER':
                        last_mix_shader = link.from_node
                        break

                if last_mix_shader:
                    links.new(last_mix_shader.outputs['Shader'], mix_shader.inputs[1])
                else:
                    links.new(pbsdf.outputs['BSDF'], mix_shader.inputs[1])
                links.new(principled_bsdf_copy.outputs['BSDF'], mix_shader.inputs[2])

                dirt_node = nodes.new(type='ShaderNodeGroup')
                dirt_node.node_tree = my_node_group
                dirt_node.location = material_output.location + Vector([1400, -250])
                links.new(dirt_node.outputs['Value'], mix_shader.inputs[0])

                links.new(mix_shader.outputs['Shader'], material_output.inputs['Surface'])
                    
        for obj in selected_objects:
            for slot in obj.material_slots:
                if slot.material and slot.material.use_nodes:
                    add_dirt_attributes_to_material(slot.material)    

        return {'FINISHED'}
        
class OBJECT_OT_DeleteDirtButton(bpy.types.Operator):
    bl_idname = "object.delete_dirt_operator"
    bl_label = "Delete All"
    bl_description = "Click this button to delete dirt"

    def execute(self, context):
        for material in bpy.data.materials:
            if not material.use_nodes:
                continue

            nodes = material.node_tree.nodes
            links = material.node_tree.links

            for node_name in ["Dirt_Musgrave_Texture", "Dirt_Noise_Texture", "Dirt_Color_Ramp"]:
                node = nodes.get(node_name)
                if node:
                    nodes.remove(node)

            dirt_bsdf = nodes.get('Dirt_Principled_BSDF')
            dirt_group = None
            for node in nodes:
                if node.type == 'GROUP' and node.node_tree and node.node_tree.name == 'Dirt':
                    dirt_group = node
                    break

            if not dirt_bsdf or not dirt_group:
                continue

            mix_shader = None
            for link in links:
                if link.from_node == dirt_bsdf and link.to_node.type == 'MIX_SHADER':
                    mix_shader = link.to_node
                    break

            if not mix_shader:
                continue

            last_shader_node = None
            for link in links:
                if link.to_node == mix_shader:
                    last_shader_node = link.from_node
                    break

            material_output = next((node for node in nodes if node.type == 'OUTPUT_MATERIAL'), None)
            if not material_output:
                continue

            if last_shader_node:
                output_socket_name = 'Shader' if last_shader_node.type == 'MIX_SHADER' else 'BSDF'
                links.new(last_shader_node.outputs[output_socket_name], material_output.inputs['Surface'])

            nodes.remove(mix_shader)
            nodes.remove(dirt_bsdf)
            nodes.remove(dirt_group)

        context.scene.aged_fx.general_dirt_scale = 0.0
        context.scene.aged_fx.edge_dirt_scale = 0.0
        context.scene.aged_fx.general_dirt = 0.0
        context.scene.aged_fx.edge_dirt = 0.0
        context.scene.aged_fx.edge_dirt_spread = 0.0
        context.scene.aged_fx.dirt_color_stop_0 = colorsys.hsv_to_rgb(0.050, 1.000, 0.050) + (1.0,)
        context.scene.aged_fx.dirt_color_stop_1 = colorsys.hsv_to_rgb(0.050, 1.000, 0.150) + (1.0,)
        context.scene.aged_fx.dirt_set = False

        objects = bpy.data.objects
        for obj in objects:
            for slot in obj.material_slots:
                if slot.material and slot.material.use_nodes:
                    remove_dirt_attributes_from_material(slot.material)

        for material in bpy.data.materials:
            self.update_materials_after_dirt_deletion(material)

        return {'FINISHED'}

    def update_materials_after_dirt_deletion(self, material):
        if material.use_nodes:
            nodes = material.node_tree.nodes
            links = material.node_tree.links

            metal_mix_shader = None
            principled_bsdf = None
            for node in nodes:
                if node.type == 'MIX_SHADER' and node.name == 'Metal_Mix_Shader':
                    metal_mix_shader = node
                elif node.type == 'BSDF_PRINCIPLED' and node.name != 'Metal_Principled_BSDF':
                    principled_bsdf = node

            material_output = next((node for node in nodes if node.type == 'OUTPUT_MATERIAL'), None)
            if metal_mix_shader and principled_bsdf and material_output:
                links.new(metal_mix_shader.inputs[1], principled_bsdf.outputs['BSDF'])
                links.new(metal_mix_shader.outputs['Shader'], material_output.inputs['Surface'])
            elif principled_bsdf and material_output:
                links.new(principled_bsdf.outputs['BSDF'], material_output.inputs['Surface'])
        
class OBJECT_OT_DeleteDirtSelectedButton(bpy.types.Operator):
    bl_idname = "object.delete_dirt_selected_operator"
    bl_label = "Delete Selected"
    bl_description = "Click this button to delete dirt from selected objects"

    def execute(self, context):
        selected_objects = [obj for obj in context.selected_objects if obj.type == 'MESH']
        for obj in selected_objects:
            for material_slot in obj.material_slots:
                material = material_slot.material
                if not material or not material.use_nodes:
                    continue

                nodes = material.node_tree.nodes
                links = material.node_tree.links

                for node_name in ["Dirt_Musgrave_Texture", "Dirt_Noise_Texture", "Dirt_Color_Ramp"]:
                    node = nodes.get(node_name)
                    if node:
                        nodes.remove(node)

                dirt_bsdf = nodes.get('Dirt_Principled_BSDF')
                dirt_group = None
                for node in nodes:
                    if node.type == 'GROUP' and node.node_tree and node.node_tree.name == 'Dirt':
                        dirt_group = node
                        break

                if dirt_bsdf:
                    mix_shader = None
                    for link in links:
                        if link.from_node == dirt_bsdf and link.to_node.type == 'MIX_SHADER':
                            mix_shader = link.to_node
                            break

                    last_shader_node = None
                    if mix_shader:
                        for link in links:
                            if link.to_node == mix_shader:
                                last_shader_node = link.from_node
                                break

                        material_output = next((node for node in nodes if node.type == 'OUTPUT_MATERIAL'), None)
                        if material_output and last_shader_node:
                            output_socket_name = 'Shader' if last_shader_node.type == 'MIX_SHADER' else 'BSDF'
                            links.new(last_shader_node.outputs[output_socket_name], material_output.inputs['Surface'])
                            nodes.remove(mix_shader)

                    elif dirt_group:
                        material_output = next((node for node in nodes if node.type == 'OUTPUT_MATERIAL'), None)
                        if material_output:
                            links.new(dirt_bsdf.outputs['BSDF'], material_output.inputs['Surface'])

                    nodes.remove(dirt_bsdf)
                    if dirt_group:
                        nodes.remove(dirt_group)

        for obj in context.selected_objects:
            for slot in obj.material_slots:
                if slot.material and slot.material.use_nodes:
                    remove_dirt_attributes_from_material(slot.material)

        for obj in context.selected_objects:
            for slot in obj.material_slots:
                if slot.material:
                    self.update_materials_after_dirt_deletion(slot.material)

        return {'FINISHED'}

    def update_materials_after_dirt_deletion(self, material):
        if material.use_nodes:
            nodes = material.node_tree.nodes
            links = material.node_tree.links

            metal_mix_shader = None
            principled_bsdf = None
            for node in nodes:
                if node.type == 'MIX_SHADER' and node.name == 'Metal_Mix_Shader':
                    metal_mix_shader = node
                elif node.type == 'BSDF_PRINCIPLED' and node.name != 'Metal_Principled_BSDF':
                    principled_bsdf = node

            material_output = next((node for node in nodes if node.type == 'OUTPUT_MATERIAL'), None)
            if metal_mix_shader and principled_bsdf and material_output:
                links.new(metal_mix_shader.inputs[1], principled_bsdf.outputs['BSDF'])
                links.new(metal_mix_shader.outputs['Shader'], material_output.inputs['Surface'])
            elif principled_bsdf and material_output:
                links.new(principled_bsdf.outputs['BSDF'], material_output.inputs['Surface'])
 
class OBJECT_OT_MakeUniqueDirt(bpy.types.Operator):
    bl_idname = "object.make_unique_dirt"
    bl_label = "Make Unique"
    bl_description = "Remove drivers for dirt properties on selected objects"

    def execute(self, context):
        for obj in context.selected_objects:
            if obj.type == 'MESH':
                for slot in obj.material_slots:
                    material = slot.material
                    if material and material.use_nodes:
                        for node in material.node_tree.nodes:
                            if node.type == 'VALUE' and node.label in ["General Dirt Scale", "Edge Dirt Scale", "General Dirt", "Edge Dirt", "Edge Dirt Spread"]:
                                data_path = f'nodes["{node.name}"].outputs[0].default_value'
                                if material.node_tree.animation_data and material.node_tree.animation_data.drivers:
                                    for driver in material.node_tree.animation_data.drivers:
                                        if driver.data_path == data_path:
                                            material.node_tree.driver_remove(driver.data_path)
                                            break

        return {'FINISHED'}
        
class OBJECT_OT_MakeMetalButton(bpy.types.Operator):
    bl_idname = "object.make_metal_operator"
    bl_label = "Add All"
    bl_description = "Click this button to add metal"

    def execute(self, context):
        addon_prefs = context.preferences.addons[__package__].preferences
        filepath = addon_prefs.filepath

        my_node_group = bpy.data.node_groups.get('Metal', None)
        if not my_node_group:
            with bpy.data.libraries.load(filepath) as (data_from, data_to):
                data_to.node_groups = ['Metal']
            my_node_group = bpy.data.node_groups.get('Metal', None)
            
            if not my_node_group:
                self.report({'ERROR'}, "The 'Metal' node group could not be imported. Please ensure the path is correct in the Addon Preferences.")
                return {'CANCELLED'}

        if not context.scene.aged_fx.metal_set:
            context.scene.aged_fx.general_metal_scale = 0.2
            context.scene.aged_fx.edge_metal_scale = 2.0
            context.scene.aged_fx.general_metal = 8.0
            context.scene.aged_fx.edge_metal = 4.0
            context.scene.aged_fx.edge_metal_spread = 0.1
            context.scene.aged_fx.metal_set = True

        excluded_collections = {"Moths AgedFX", "Ants AgedFX", "Spiders AgedFX"}
        objects = set(obj for obj in bpy.data.objects if obj.type == 'MESH' and not is_object_in_excluded_collections(obj, excluded_collections))
        materials_done = set()

        for obj in objects:
            for slot in obj.material_slots:
                material = slot.material
                if not material or not material.use_nodes:
                    continue

                if material.name == "Cobweb Generator":
                    continue

                if any(node for node in material.node_tree.nodes if node.type == 'GROUP' and node.node_tree == my_node_group):
                    continue

                if material.name in materials_done:
                    continue

                nodes = material.node_tree.nodes
                links = material.node_tree.links

                pbsdf = find_original_principled_bsdf(material)
                if pbsdf is None:
                    continue

                material_output = next((node for node in nodes if node.type == 'OUTPUT_MATERIAL'), None)
                if material_output is None:
                    continue

                last_mix_shader = None
                for link in links:
                    if link.to_node == material_output and link.from_node.type == 'MIX_SHADER':
                        last_mix_shader = link.from_node
                        break

                noise_texture = nodes.new(type='ShaderNodeTexNoise')
                noise_texture.location = material_output.location + Vector([1600, -650])
                noise_texture.inputs['Scale'].default_value = 100
                noise_texture.inputs['Detail'].default_value = 15
                noise_texture.inputs['Roughness'].default_value = 1
                noise_texture.inputs['Lacunarity'].default_value = 2
                noise_texture.inputs['Distortion'].default_value = 0
                noise_texture.name = "Metal_Noise_Texture"

                color_ramp = nodes.new(type='ShaderNodeValToRGB')
                color_ramp.location = material_output.location + Vector([1800, -650])
                color_ramp.color_ramp.elements[0].color = colorsys.hsv_to_rgb(0.000, 0.000, 0.400) + (1.0,)
                color_ramp.color_ramp.elements[1].color = colorsys.hsv_to_rgb(0.000, 0.000, 0.600) + (1.0,)
                color_ramp.color_ramp.elements[0].position = 0.4
                color_ramp.color_ramp.elements[1].position = 0.6
                color_ramp.name = "Metal_Color_Ramp"

                links.new(noise_texture.outputs['Fac'], color_ramp.inputs['Fac'])

                mix_shader = nodes.new(type='ShaderNodeMixShader')
                mix_shader.location = material_output.location + Vector([2400, -650])
                mix_shader.name = "Metal_Mix_Shader"

                principled_bsdf_copy = nodes.new(type='ShaderNodeBsdfPrincipled')
                principled_bsdf_copy.location = material_output.location + Vector([2100, -650])
                principled_bsdf_copy.name = "Metal_Principled_BSDF"
                links.new(color_ramp.outputs['Color'], principled_bsdf_copy.inputs['Base Color'])
                
                principled_bsdf_copy.inputs['Metallic'].default_value = 1
                principled_bsdf_copy.inputs['Roughness'].default_value = 0.05

                if last_mix_shader:
                    links.new(last_mix_shader.outputs['Shader'], mix_shader.inputs[1])
                else:
                    links.new(pbsdf.outputs['BSDF'], mix_shader.inputs[1])
                links.new(principled_bsdf_copy.outputs['BSDF'], mix_shader.inputs[2])

                metal_node = nodes.new(type='ShaderNodeGroup')
                metal_node.node_tree = my_node_group
                metal_node.location = material_output.location + Vector([1400, -650])
                links.new(metal_node.outputs['Value'], mix_shader.inputs[0])

                links.new(mix_shader.outputs['Shader'], material_output.inputs['Surface'])

                materials_done.add(material.name)

        for obj in bpy.data.objects:
            for slot in obj.material_slots:
                if slot.material and slot.material.use_nodes:
                    add_metal_attributes_to_material(slot.material)

        return {'FINISHED'}

class OBJECT_OT_MakeMetalSelectedButton(bpy.types.Operator):
    bl_idname = "object.make_metal_selected_operator"
    bl_label = "Add Selected"
    bl_description = "Click this button to add metal to selected materials"

    def execute(self, context):
        addon_prefs = context.preferences.addons[__package__].preferences
        filepath = addon_prefs.filepath

        my_node_group = bpy.data.node_groups.get('Metal', None)
        if not my_node_group:
            with bpy.data.libraries.load(filepath) as (data_from, data_to):
                data_to.node_groups = ['Metal']
            my_node_group = bpy.data.node_groups.get('Metal', None)

            if not my_node_group:
                self.report({'ERROR'}, "The 'Metal' node group could not be imported. Please ensure the path is correct in the Addon Preferences.")
                return {'CANCELLED'}

        if not context.scene.aged_fx.metal_set:
            context.scene.aged_fx.general_metal_scale = 0.2
            context.scene.aged_fx.edge_metal_scale = 2.0
            context.scene.aged_fx.general_metal = 8.0
            context.scene.aged_fx.edge_metal = 4.0
            context.scene.aged_fx.edge_metal_spread = 0.1
            context.scene.aged_fx.metal_color_stop_0 = colorsys.hsv_to_rgb(0.000, 0.000, 0.400) + (1.0,)
            context.scene.aged_fx.metal_color_stop_1 = colorsys.hsv_to_rgb(0.000, 0.000, 0.600) + (1.0,)
            context.scene.aged_fx.metal_set = True

        selected_objects = [obj for obj in context.selected_objects if obj.type == 'MESH']
        for obj in selected_objects:
            for mat_slot in obj.material_slots:
                material = mat_slot.material
                if not material or not material.use_nodes:
                    continue

                if material.name == "Cobweb Generator":
                    continue

                if any(node for node in material.node_tree.nodes if node.type == 'GROUP' and node.node_tree == my_node_group):
                    continue

                nodes = material.node_tree.nodes
                links = material.node_tree.links

                pbsdf = find_original_principled_bsdf(material)
                if pbsdf is None:
                    continue

                material_output = next((node for node in nodes if node.type == 'OUTPUT_MATERIAL'), None)
                if material_output is None:
                    continue

                noise_texture = nodes.new(type='ShaderNodeTexNoise')
                noise_texture.location = material_output.location + Vector([1600, -650])
                noise_texture.inputs['Scale'].default_value = 100
                noise_texture.inputs['Detail'].default_value = 15
                noise_texture.inputs['Roughness'].default_value = 1
                noise_texture.inputs['Lacunarity'].default_value = 2
                noise_texture.inputs['Distortion'].default_value = 0
                noise_texture.name = "Metal_Noise_Texture"

                color_ramp = nodes.new(type='ShaderNodeValToRGB')
                color_ramp.location = material_output.location + Vector([1800, -650])
                color_ramp.color_ramp.elements[0].color = colorsys.hsv_to_rgb(0.000, 0.000, 0.800) + (1.0,)
                color_ramp.color_ramp.elements[1].color = colorsys.hsv_to_rgb(0.000, 0.000, 0.900) + (1.0,)
                color_ramp.color_ramp.elements[0].position = 0.4
                color_ramp.color_ramp.elements[1].position = 0.6
                color_ramp.name = "Metal_Color_Ramp"

                links.new(noise_texture.outputs['Fac'], color_ramp.inputs['Fac'])

                mix_shader = nodes.new(type='ShaderNodeMixShader')
                mix_shader.location = material_output.location + Vector([2400, -650])
                mix_shader.name = "Metal_Mix_Shader"

                principled_bsdf_copy = nodes.new(type='ShaderNodeBsdfPrincipled')
                principled_bsdf_copy.location = material_output.location + Vector([2100, -650])
                principled_bsdf_copy.name = "Metal_Principled_BSDF"
                links.new(color_ramp.outputs['Color'], principled_bsdf_copy.inputs['Base Color'])
                
                principled_bsdf_copy.inputs['Metallic'].default_value = 1
                principled_bsdf_copy.inputs['Roughness'].default_value = 0.05

                last_mix_shader = None
                for link in links:
                    if link.to_node == material_output and link.from_node.type == 'MIX_SHADER':
                        last_mix_shader = link.from_node
                        break

                if last_mix_shader:
                    links.new(last_mix_shader.outputs['Shader'], mix_shader.inputs[1])
                else:
                    links.new(pbsdf.outputs['BSDF'], mix_shader.inputs[1])
                links.new(principled_bsdf_copy.outputs['BSDF'], mix_shader.inputs[2])

                metal_node = nodes.new(type='ShaderNodeGroup')
                metal_node.node_tree = my_node_group
                metal_node.location = material_output.location + Vector([1400, -650])
                links.new(metal_node.outputs['Value'], mix_shader.inputs[0])

                links.new(mix_shader.outputs['Shader'], material_output.inputs['Surface'])

        for obj in selected_objects:
            for slot in obj.material_slots:
                if slot.material and slot.material.use_nodes:
                    add_metal_attributes_to_material(slot.material)

        return {'FINISHED'}

class OBJECT_OT_DeleteMetalButton(bpy.types.Operator):
    bl_idname = "object.delete_metal_operator"
    bl_label = "Delete All"
    bl_description = "Click this button to delete metal"

    def execute(self, context):
        for material in bpy.data.materials:
            if not material.use_nodes:
                continue

            nodes = material.node_tree.nodes
            links = material.node_tree.links

            for node_name in ["Metal_Noise_Texture", "Metal_Color_Ramp"]:
                node = nodes.get(node_name)
                if node:
                    nodes.remove(node)

            metal_bsdf = nodes.get('Metal_Principled_BSDF')
            metal_group = None
            for node in nodes:
                if node.type == 'GROUP' and node.node_tree and node.node_tree.name == 'Metal':
                    metal_group = node
                    break

            if not metal_bsdf or not metal_group:
                continue

            mix_shader = None
            for link in links:
                if link.from_node == metal_bsdf and link.to_node.type == 'MIX_SHADER':
                    mix_shader = link.to_node
                    break

            if not mix_shader:
                continue

            last_shader_node = None
            for link in links:
                if link.to_node == mix_shader:
                    last_shader_node = link.from_node
                    break

            material_output = next((node for node in nodes if node.type == 'OUTPUT_MATERIAL'), None)
            if not material_output:
                continue

            if last_shader_node:
                output_socket_name = 'Shader' if last_shader_node.type == 'MIX_SHADER' else 'BSDF'
                links.new(last_shader_node.outputs[output_socket_name], material_output.inputs['Surface'])

            nodes.remove(mix_shader)
            nodes.remove(metal_bsdf)
            nodes.remove(metal_group)

        context.scene.aged_fx.general_metal_scale = 0.0
        context.scene.aged_fx.edge_metal_scale = 0.0
        context.scene.aged_fx.general_metal = 0.0
        context.scene.aged_fx.edge_metal = 0.0
        context.scene.aged_fx.edge_metal_spread = 0.0
        context.scene.aged_fx.metal_set = False

        objects = bpy.data.objects
        for obj in objects:
            for slot in obj.material_slots:
                if slot.material and slot.material.use_nodes:
                    remove_metal_attributes_from_material(slot.material)

        for material in bpy.data.materials:
            self.update_materials_after_metal_deletion(material)

        return {'FINISHED'}

    def update_materials_after_metal_deletion(self, material):
        if material.use_nodes:
            nodes = material.node_tree.nodes
            links = material.node_tree.links

            dirt_mix_shader = None
            principled_bsdf = None
            for node in nodes:
                if node.type == 'MIX_SHADER' and node.name == 'Dirt_Mix_Shader':
                    dirt_mix_shader = node
                elif node.type == 'BSDF_PRINCIPLED' and node.name != 'Dirt_Principled_BSDF':
                    principled_bsdf = node

            material_output = next((node for node in nodes if node.type == 'OUTPUT_MATERIAL'), None)
            if dirt_mix_shader and principled_bsdf and material_output:
                links.new(dirt_mix_shader.inputs[1], principled_bsdf.outputs['BSDF'])
                links.new(dirt_mix_shader.outputs['Shader'], material_output.inputs['Surface'])
            elif principled_bsdf and material_output:
                links.new(principled_bsdf.outputs['BSDF'], material_output.inputs['Surface'])

class OBJECT_OT_DeleteMetalSelectedButton(bpy.types.Operator):
    bl_idname = "object.delete_metal_selected_operator"
    bl_label = "Delete Selected"
    bl_description = "Click this button to delete metal from selected objects"

    def execute(self, context):
        selected_objects = [obj for obj in context.selected_objects if obj.type == 'MESH']
        for obj in selected_objects:
            for material_slot in obj.material_slots:
                material = material_slot.material
                if not material or not material.use_nodes:
                    continue

                nodes = material.node_tree.nodes
                links = material.node_tree.links

                for node_name in ["Metal_Noise_Texture", "Metal_Color_Ramp"]:
                    node = nodes.get(node_name)
                    if node:
                        nodes.remove(node)

                metal_bsdf = nodes.get('Metal_Principled_BSDF')
                metal_group = None
                for node in nodes:
                    if node.type == 'GROUP' and node.node_tree and node.node_tree.name == 'Metal':
                        metal_group = node
                        break

                if metal_bsdf:
                    mix_shader = None
                    for link in links:
                        if link.from_node == metal_bsdf and link.to_node.type == 'MIX_SHADER':
                            mix_shader = link.to_node
                            break

                    last_shader_node = None
                    if mix_shader:
                        for link in links:
                            if link.to_node == mix_shader:
                                last_shader_node = link.from_node
                                break

                        material_output = next((node for node in nodes if node.type == 'OUTPUT_MATERIAL'), None)
                        if material_output and last_shader_node:
                            output_socket_name = 'Shader' if last_shader_node.type == 'MIX_SHADER' else 'BSDF'
                            links.new(last_shader_node.outputs[output_socket_name], material_output.inputs['Surface'])
                            nodes.remove(mix_shader)

                    elif metal_group:
                        material_output = next((node for node in nodes if node.type == 'OUTPUT_MATERIAL'), None)
                        if material_output:
                            links.new(metal_bsdf.outputs['BSDF'], material_output.inputs['Surface'])

                    nodes.remove(metal_bsdf)
                    if metal_group:
                        nodes.remove(metal_group)

        for obj in context.selected_objects:
            for slot in obj.material_slots:
                if slot.material and slot.material.use_nodes:
                    remove_metal_attributes_from_material(slot.material)

        for obj in context.selected_objects:
            for slot in obj.material_slots:
                if slot.material:
                    self.update_materials_after_metal_deletion(slot.material)

        return {'FINISHED'}

    def update_materials_after_metal_deletion(self, material):
        if material.use_nodes:
            nodes = material.node_tree.nodes
            links = material.node_tree.links

            dirt_mix_shader = None
            principled_bsdf = None
            for node in nodes:
                if node.type == 'MIX_SHADER' and node.name == 'Dirt_Mix_Shader':
                    dirt_mix_shader = node
                elif node.type == 'BSDF_PRINCIPLED' and node.name != 'Dirt_Principled_BSDF':
                    principled_bsdf = node

            material_output = next((node for node in nodes if node.type == 'OUTPUT_MATERIAL'), None)
            if dirt_mix_shader and principled_bsdf and material_output:
                links.new(dirt_mix_shader.inputs[1], principled_bsdf.outputs['BSDF'])
                links.new(dirt_mix_shader.outputs['Shader'], material_output.inputs['Surface'])
            elif principled_bsdf and material_output:
                links.new(principled_bsdf.outputs['BSDF'], material_output.inputs['Surface'])

class OBJECT_OT_MakeUniqueMetal(bpy.types.Operator):
    bl_idname = "object.make_unique_metal"
    bl_label = "Make Unique"
    bl_description = "Remove drivers for metal properties on selected objects"

    def execute(self, context):
        for obj in context.selected_objects:
            if obj.type == 'MESH':
                for slot in obj.material_slots:
                    material = slot.material
                    if material and material.use_nodes:
                        for node in material.node_tree.nodes:
                            if node.type == 'VALUE' and node.label in ["General Metal Scale", "Edge Metal Scale", "General Metal", "Edge Metal", "Edge Metal Spread"]:
                                data_path = f'nodes["{node.name}"].outputs[0].default_value'
                                if material.node_tree.animation_data and material.node_tree.animation_data.drivers:
                                    for driver in material.node_tree.animation_data.drivers:
                                        if driver.data_path == data_path:
                                            material.node_tree.driver_remove(driver.data_path)
                                            break

        return {'FINISHED'}
        
class OBJECT_OT_MakeSelectedRustyButton(bpy.types.Operator):
    bl_idname = "object.make_selected_rusty"
    bl_label = "Make Selected Rusty"

    def execute(self, context):
        selected_objects = context.selected_objects
        for obj in selected_objects:
            if obj.type == 'MESH' and obj.material_slots:
                for slot in obj.material_slots:
                    material = slot.material
                    if material and material.use_nodes:
                        color_ramp_node = material.node_tree.nodes.get('Metal_Color_Ramp')
                        if color_ramp_node:
                            color_ramp_node.color_ramp.elements[0].color = colorsys.hsv_to_rgb(0.050, 1.000, 0.050) + (1.0,)
                            color_ramp_node.color_ramp.elements[1].color = colorsys.hsv_to_rgb(0.050, 1.000, 0.300) + (1.0,)
                        bsdf_node = material.node_tree.nodes.get('Metal_Principled_BSDF')
                        if bsdf_node:
                            bsdf_node.inputs['Roughness'].default_value = 0.5
        return {'FINISHED'}

class OBJECT_OT_UndoRustyButton(bpy.types.Operator):
    bl_idname = "object.undo_rusty"
    bl_label = "Undo"

    def execute(self, context):
        selected_objects = context.selected_objects
        for obj in selected_objects:
            if obj.type == 'MESH' and obj.material_slots:
                for slot in obj.material_slots:
                    material = slot.material
                    if material and material.use_nodes:
                        color_ramp_node = material.node_tree.nodes.get('Metal_Color_Ramp')
                        if color_ramp_node:
                            color_ramp_node.color_ramp.elements[0].color = colorsys.hsv_to_rgb(0.000, 0.000, 0.400) + (1.0,)
                            color_ramp_node.color_ramp.elements[1].color = colorsys.hsv_to_rgb(0.000, 0.000, 0.600) + (1.0,)
                        bsdf_node = material.node_tree.nodes.get('Metal_Principled_BSDF')
                        if bsdf_node:
                            bsdf_node.inputs['Roughness'].default_value = 0.05
        return {'FINISHED'}

class OBJECT_OT_MakeDeformationsButton(bpy.types.Operator):
    bl_idname = "object.make_deformations_operator"
    bl_label = "Add All"
    bl_description = "Click this button to add deformations"

    def execute(self, context):
        addon_prefs = context.preferences.addons[__package__].preferences
        filepath = addon_prefs.filepath

        my_node_group = bpy.data.node_groups.get('Deformations', None)
        if not my_node_group:
            with bpy.data.libraries.load(filepath) as (data_from, data_to):
                data_to.node_groups = ['Deformations']
            my_node_group = bpy.data.node_groups.get('Deformations', None)
            
            if not my_node_group:
                self.report({'ERROR'}, "The node group 'Deformations' could not be imported. Please ensure the path in the Addon Preferences is correct.")
                return {'CANCELLED'}

        if not context.scene.aged_fx.deformations_set:
            context.scene.aged_fx.deformations_amount = 0.5
            context.scene.aged_fx.deformations_scale = 8.0
            context.scene.aged_fx.scratches = 0.4
            context.scene.aged_fx.bumps = 0.2
            context.scene.aged_fx.deformations_set = True

        excluded_collections = {"Moths AgedFX", "Ants AgedFX", "Spiders AgedFX"}
        materials_done = set()

        for obj in bpy.data.objects:
            if obj.type == 'MESH' and not is_object_in_excluded_collections(obj, excluded_collections):
                for slot in obj.material_slots:
                    if slot.material is not None and slot.material.use_nodes:
                        add_deformations_to_material(slot.material, context.scene.aged_fx.deformations_amount, materials_done)

        objects = bpy.data.objects
        for obj in objects:
            for slot in obj.material_slots:
                if slot.material and slot.material.use_nodes:
                    add_deformations_attributes_to_material(slot.material)

        return {'FINISHED'}
        
class OBJECT_OT_MakeDeformationsSelectedButton(Operator):
    bl_idname = "object.make_deformations_selected_operator"
    bl_label = "Add Selected"
    bl_description = "Click this button to add deformations to selected objects"

    def execute(self, context):
        addon_prefs = context.preferences.addons[__package__].preferences
        filepath = addon_prefs.filepath

        my_node_group = bpy.data.node_groups.get('Deformations', None)
        if not my_node_group:
            with bpy.data.libraries.load(filepath) as (data_from, data_to):
                data_to.node_groups = ['Deformations']
            my_node_group = bpy.data.node_groups.get('Deformations', None)
            
            if not my_node_group:
                self.report({'ERROR'}, "Die Node-Gruppe 'Deformations' konnte nicht importiert werden. Stellen Sie sicher, dass der Pfad in den Addon Preferences korrekt ist.")
                return {'CANCELLED'}

        materials_done = set()

        selected_objects = [obj for obj in context.selected_objects if obj.type == 'MESH']

        for obj in selected_objects:
            for slot in obj.material_slots:
                if slot.material is not None and slot.material.use_nodes:
                    add_deformations_to_material(slot.material, context.scene.aged_fx.deformations_amount, materials_done)

        if not context.scene.aged_fx.deformations_set:
            context.scene.aged_fx.deformations_amount = 0.5
            context.scene.aged_fx.deformations_scale = 8.0
            context.scene.aged_fx.scratches = 0.4
            context.scene.aged_fx.bumps = 0.2
            context.scene.aged_fx.deformations_set = True
            
        selected_objects = context.selected_objects
        for obj in selected_objects:
            for slot in obj.material_slots:
                if slot.material and slot.material.use_nodes:
                    add_deformations_attributes_to_material(slot.material)

        return {'FINISHED'}

class OBJECT_OT_DeleteDeformationsButton(Operator):
    bl_idname = "object.delete_deformations_operator"
    bl_label = "Delete All"
    bl_description = "Click this button to delete deformations"

    def execute(self, context):
        for obj in bpy.data.objects:
            if obj.type == 'MESH':
                for slot in obj.material_slots:
                    if slot.material is not None and slot.material.use_nodes:
                        remove_deformations_from_material(slot.material)

        context.scene.aged_fx.deformations_amount = 0.0
        context.scene.aged_fx.deformations_scale = 0.0
        context.scene.aged_fx.scratches = 0.0
        context.scene.aged_fx.bumps = 0.0
        context.scene.aged_fx.deformations_set = False
        
        objects = bpy.data.objects
        for obj in objects:
            for slot in obj.material_slots:
                if slot.material and slot.material.use_nodes:
                    remove_deformations_attributes_from_material(slot.material)

        return {'FINISHED'}
        
class OBJECT_OT_DeleteDeformationsSelectedButton(Operator):
    bl_idname = "object.delete_deformations_selected_operator"
    bl_label = "Delete Selected"
    bl_description = "Click this button to delete deformations from selected objects"

    def execute(self, context):
        delete_deformations_selected(self, context)
        
        selected_objects = context.selected_objects
        for obj in selected_objects:
            for slot in obj.material_slots:
                if slot.material and slot.material.use_nodes:
                    remove_deformations_attributes_from_material(slot.material)
                                        
        return {'FINISHED'}
        
class OBJECT_OT_MakeUniqueDeformations(bpy.types.Operator):
    bl_idname = "object.make_unique_deformations"
    bl_label = "Make Unique"
    bl_description = "Remove drivers for deformations properties on selected objects"

    def execute(self, context):
        for obj in context.selected_objects:
            if obj.type == 'MESH':
                for slot in obj.material_slots:
                    material = slot.material
                    if material and material.use_nodes:
                        for node in material.node_tree.nodes:
                            if node.type == 'VALUE' and node.label in ["Deformations Amount", "Deformations Scale", "Scratches", "Bumps"]:
                                data_path = f'nodes["{node.name}"].outputs[0].default_value'
                                if material.node_tree.animation_data and material.node_tree.animation_data.drivers:
                                    for driver in material.node_tree.animation_data.drivers:
                                        if driver.data_path == data_path:
                                            material.node_tree.driver_remove(driver.data_path)
                                            break

        return {'FINISHED'}
        
class OBJECT_OT_AddCobweb(bpy.types.Operator):
    bl_idname = "object.add_cobweb"
    bl_label = "Add Cobweb"
    bl_description = "Add a Cobweb Generator"

    def execute(self, context):
        addon_prefs = context.preferences.addons[__package__].preferences
        cobweb_collection_name = "Cobweb Generator"
        just_imported = False

        if cobweb_collection_name not in bpy.data.collections:
            bpy.ops.wm.append(
                filename=cobweb_collection_name,
                directory=os.path.join(addon_prefs.filepath, "Collection")
            )
            just_imported = True
            context.scene.aged_fx.cobweb_translucency = 0.5

        cobweb_collection = bpy.data.collections.get(cobweb_collection_name)
        
        if cobweb_collection and cobweb_collection.objects:
            if just_imported:
                cobweb_obj = cobweb_collection.objects[0]
                context.view_layer.objects.active = cobweb_obj
                bpy.ops.object.select_all(action='DESELECT')
                cobweb_obj.select_set(True)

                if cobweb_obj.type == 'CURVE':
                    bpy.ops.object.mode_set(mode='EDIT')

                    bpy.ops.wm.tool_set_by_id(name="builtin.draw")
                    
                    for area in context.screen.areas:
                        if area.type == 'PROPERTIES':
                            for space in area.spaces:
                                if space.type == 'PROPERTIES':
                                    try:
                                        space.context = 'MODIFIER'
                                    except TypeError:
                                        print("Modifier context not available in this area.")
                    
                    tool_settings = context.tool_settings
                    if tool_settings and hasattr(tool_settings, 'curve_paint_settings'):
                        curve_settings = tool_settings.curve_paint_settings
                        curve_settings.depth_mode = 'SURFACE'
            else:
                first_obj = cobweb_collection.objects[0]
                context.view_layer.objects.active = first_obj
                bpy.ops.object.select_all(action='DESELECT')
                first_obj.select_set(True)
                bpy.ops.object.duplicate()
                new_obj = context.view_layer.objects.active

                new_obj.name = "Cobweb Generator"

                bpy.ops.object.mode_set(mode='EDIT')

                if new_obj.type == 'CURVE':
                    bpy.ops.curve.select_all(action='SELECT')
                    bpy.ops.curve.delete(type='VERT')

                    bpy.ops.wm.tool_set_by_id(name="builtin.draw")
                    
                    for area in context.screen.areas:
                        if area.type == 'PROPERTIES':
                            for space in area.spaces:
                                if space.type == 'PROPERTIES':
                                    try:
                                        space.context = 'MODIFIER'
                                    except TypeError:
                                        print("Modifier context not available in this area.")
                    
                    tool_settings = context.tool_settings
                    if tool_settings and hasattr(tool_settings, 'curve_paint_settings'):
                        curve_settings = tool_settings.curve_paint_settings
                        curve_settings.depth_mode = 'SURFACE'
        else:
            self.report({'WARNING'}, "Cobweb Generator object was not found or could not be imported correctly.")
            return {'CANCELLED'}

        return {'FINISHED'}
        
class OBJECT_OT_DeleteCobweb(Operator):
    bl_idname = "object.delete_cobweb"
    bl_label = "Delete Cobweb"
    bl_description = "Delete the selected Cobweb object"

    def execute(self, context):
        selected_obj = context.active_object
        cobweb_collection_name = "Cobweb Generator"

        if selected_obj and selected_obj.name in bpy.data.collections[cobweb_collection_name].objects:
            bpy.data.objects.remove(selected_obj, do_unlink=True)

            cobweb_collection = bpy.data.collections[cobweb_collection_name]
            if len(cobweb_collection.objects) == 0:
                bpy.data.collections.remove(cobweb_collection)
                bpy.ops.outliner.orphans_purge()

        return {'FINISHED'}
        
class OBJECT_OT_ConvertSelectedToMesh(Operator):
    bl_idname = "object.convert_selected_to_mesh"
    bl_label = "Convert Selected to Mesh"
    bl_description = "Convert selected objects to Mesh"

    def execute(self, context):
        bpy.ops.object.convert(target='MESH')
        return {'FINISHED'}

class OBJECT_OT_AddMossGenerator(bpy.types.Operator):
    bl_idname = "object.add_moss_generator"
    bl_label = "Add Moss Generator"
    bl_description = "Add Moss Generator to selected objects"

    def execute(self, context):
        addon_prefs = context.preferences.addons[__package__].preferences
        filepath = addon_prefs.filepath

        moss_node_tree_name = "Moss Generator"
        moss_node_tree = bpy.data.node_groups.get(moss_node_tree_name, None)

        if not moss_node_tree:
            with bpy.data.libraries.load(filepath, link=False) as (data_from, data_to):
                data_to.node_groups = [name for name in data_from.node_groups if name == moss_node_tree_name]
            
            moss_node_tree = bpy.data.node_groups.get(moss_node_tree_name, None)
            if not moss_node_tree:
                self.report({'ERROR'}, "The Moss Generator node tree could not be imported.")
                return {'CANCELLED'}

        apply_scale = context.scene.aged_fx.apply_scale

        for obj in context.selected_objects:
            if obj.type == 'MESH':
                if "UVMap" not in obj.data.uv_layers:
                    obj.data.uv_layers.new(name="UVMap")

                if "Moss Generator" not in obj.modifiers:
                    context.view_layer.objects.active = obj
                    bpy.ops.object.mode_set(mode='OBJECT')
                    
                    if apply_scale:
                        bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)

                    modifier = obj.modifiers.new(name="Moss Generator", type='NODES')
                    modifier.node_group = moss_node_tree
                else:
                    self.report({'INFO'}, "Moss Generator already added.")
                    
        for area in context.screen.areas:
            if area.type == 'PROPERTIES':
                for space in area.spaces:
                    if space.type == 'PROPERTIES':
                        space.context = 'MODIFIER'

        return {'FINISHED'}
        
class OBJECT_OT_RemoveMossGenerator(bpy.types.Operator):
    bl_idname = "object.remove_moss_generator"
    bl_label = "Remove Moss Generator"
    bl_description = "Remove Moss Generator from selected objects"

    def execute(self, context):
        for obj in context.selected_objects:
            if obj.type == 'MESH':
                for modifier in obj.modifiers:
                    if modifier.name == "Moss Generator" and modifier.type == 'NODES':
                        obj.modifiers.remove(modifier)
                        break

        return {'FINISHED'}
        
class OBJECT_OT_AddFog(bpy.types.Operator):
    bl_idname = "object.add_fog"
    bl_label = "Add Fog"
    bl_description = "Add Fog to the Scene"

    def execute(self, context):
        addon_prefs = context.preferences.addons[__package__].preferences
        filepath = addon_prefs.filepath

        fog_collection_name = "Fog AgedFX"
        fog_object_name = "Fog AgedFX"

        if fog_collection_name not in bpy.data.collections:
            with bpy.data.libraries.load(filepath, link=False) as (data_from, data_to):
                data_to.collections = [name for name in data_from.collections if name == fog_collection_name]

            fog_collection = bpy.data.collections.get(fog_collection_name, None)
            if not fog_collection:
                self.report({'ERROR'}, "Fog AgedFX collection not found in the specified file.")
                return {'CANCELLED'}

            if fog_collection.name not in context.scene.collection.children:
                context.scene.collection.children.link(fog_collection)
        else:
            fog_collection = bpy.data.collections[fog_collection_name]

        fog_object = next((obj for obj in fog_collection.objects if obj.name == fog_object_name), None)
        if fog_object:
            min_coords, max_coords = self.calculate_scene_bounding_box(context)
            scene_dimensions = max_coords - min_coords
            scene_center = (min_coords + max_coords) / 2

            fog_object.location = scene_center
            scale_factors = [scene_dimensions[i] / fog_object.dimensions[i] for i in range(3)]
            fog_object.scale = Vector(scale_factors)

            bpy.ops.object.select_all(action='DESELECT')
            context.view_layer.objects.active = fog_object
            fog_object.select_set(True)

            bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)
            
            bbox_corners = [Vector(corner) for corner in fog_object.bound_box]
            z_min = min(corner.z for corner in bbox_corners)
            new_origin = Vector((0, 0, z_min))
            fog_object.data.transform(Matrix.Translation(-new_origin))
            fog_object.location.z += z_min
            
            context.scene.aged_fx.fog_density = 0.05
            context.scene.aged_fx.fog_anisotropy = 0.0
            context.scene.aged_fx.fog_x_scale = 1.0
            context.scene.aged_fx.fog_y_scale = 1.0
            context.scene.aged_fx.fog_z_scale = 1.0
            context.scene.aged_fx.ground_fog_density = 0.0
            context.scene.aged_fx.ground_fog_scale = 1.0
            context.scene.aged_fx.ground_fog_height = 1.0
            context.scene.aged_fx.ground_fog_position_x = 0.0
            context.scene.aged_fx.ground_fog_position_y = 0.0
            context.scene.aged_fx.ground_fog_speed_x = 0.5
            context.scene.aged_fx.ground_fog_speed_y = 0.5
            context.scene.aged_fx.ground_fog_turbolence = 0.5
            context.scene.aged_fx.fog_color = (0.5, 0.5, 0.5, 1)

            self.add_scale_drivers(fog_object)

        else:
            self.report({'ERROR'}, "Fog AgedFX object not found in the collection.")
            return {'CANCELLED'}

        return {'FINISHED'}

    def calculate_scene_bounding_box(self, context):
        all_objects = [obj for obj in context.scene.objects if obj.type == 'MESH']
        if not all_objects:
            return Vector((0, 0, 0)), Vector((0, 0, 0))

        min_coords = Vector((float('inf'), float('inf'), float('inf')))
        max_coords = Vector((float('-inf'), float('-inf'), float('-inf')))

        for obj in all_objects:
            for vertex in obj.bound_box:
                global_vertex = obj.matrix_world @ Vector(vertex)
                min_coords = Vector(map(min, zip(min_coords, global_vertex)))
                max_coords = Vector(map(max, zip(max_coords, global_vertex)))

        return min_coords, max_coords

    def add_scale_drivers(self, fog_object):
        scale_axes = ['x', 'y', 'z']
        scale_driver_paths = [
            'aged_fx.fog_x_scale',
            'aged_fx.fog_y_scale',
            'aged_fx.fog_z_scale',
        ]

        for axis, driver_path in zip(scale_axes, scale_driver_paths):
            fcurve = fog_object.driver_add('scale', axis_map[axis])
            driver = fcurve.driver
            driver.type = 'SCRIPTED'
            driver.expression = "var"

            var = driver.variables.new()
            var.name = "var"
            var.type = 'SINGLE_PROP'
            var.targets[0].id_type = 'SCENE'
            var.targets[0].id = bpy.context.scene
            var.targets[0].data_path = driver_path

axis_map = {'x': 0, 'y': 1, 'z': 2}
        
class OBJECT_OT_DeleteFog(Operator):
    bl_idname = "object.delete_fog"
    bl_label = "Delete Fog"
    bl_description = "Delete Fog completely from the Scene"

    def execute(self, context):
        fog_collection_name = "Fog AgedFX"
        
        context.scene.aged_fx.fog_density = 0.0
        context.scene.aged_fx.fog_anisotropy = 0.0
        context.scene.aged_fx.fog_x_scale = 0.0
        context.scene.aged_fx.fog_y_scale = 0.0
        context.scene.aged_fx.fog_z_scale = 0.0
        context.scene.aged_fx.ground_fog_density = 0.0
        context.scene.aged_fx.ground_fog_scale = 0.0
        context.scene.aged_fx.ground_fog_height = 0.0
        context.scene.aged_fx.ground_fog_position_x = 0.0
        context.scene.aged_fx.ground_fog_position_y = 0.0
        context.scene.aged_fx.ground_fog_speed_x = 0.0
        context.scene.aged_fx.ground_fog_speed_y = 0.0
        context.scene.aged_fx.ground_fog_turbolence = 0.0
        context.scene.aged_fx.fog_color = (0.5, 0.5, 0.5, 1)

        if fog_collection_name in bpy.data.collections:
            fog_collection = bpy.data.collections[fog_collection_name]

            for obj in fog_collection.objects:
                bpy.data.objects.remove(obj, do_unlink=True)

            bpy.data.collections.remove(fog_collection)

            for block in bpy.data.meshes:
                if block.users == 0:
                    bpy.data.meshes.remove(block)

        return {'FINISHED'}
                                
class OBJECT_OT_AddDustParticles(Operator):
    bl_idname = "object.add_dust_particles"
    bl_label = "Add Dust Particles"
    bl_description = "Add Dust Particles to the Scene"

    def execute(self, context):
        addon_prefs = context.preferences.addons[__package__].preferences
        filepath = addon_prefs.filepath

        dust_particles_collection_name = "Dust Particles"
        
        context.scene.aged_fx.dust_particles_size_x = 10.0
        context.scene.aged_fx.dust_particles_size_y = 10.0
        context.scene.aged_fx.dust_particles_size_z = 5.0
        context.scene.aged_fx.dust_particles_wind_x = 0.0
        context.scene.aged_fx.dust_particles_wind_y = 0.0
        context.scene.aged_fx.dust_particles_wind_z = 0.0
        context.scene.aged_fx.dust_particles_amount = 100.0
        context.scene.aged_fx.dust_particles_scale_min = 0.1
        context.scene.aged_fx.dust_particles_scale_max = 1.0
        context.scene.aged_fx.dust_particles_speed = 5.0
        context.scene.aged_fx.dust_particles_rotation = 0.5
        context.scene.aged_fx.dust_particles_turbolence = 0.0
        context.scene.aged_fx.dust_particles_seed = 0
        context.scene.aged_fx.dust_particles_emission = 0.0

        if dust_particles_collection_name in bpy.data.collections:
            self.report({'INFO'}, "Dust Particles Collection is already in the scene")
            return {'FINISHED'}

        try:
            with bpy.data.libraries.load(filepath, link=False) as (data_from, data_to):
                data_to.collections = [name for name in data_from.collections if name == dust_particles_collection_name]
            
            imported_collection = bpy.data.collections.get(dust_particles_collection_name)
            if imported_collection:
                context.scene.collection.children.link(imported_collection)

                dust_particles_obj = None
                for obj in imported_collection.objects:
                    if obj.name == "Dust Particles":
                        dust_particles_obj = obj
                        break

                if dust_particles_obj:
                    active_camera = context.scene.camera
                    if active_camera:
                        dust_particles_obj.location = active_camera.location
                    else:
                        dust_particles_obj.location = (0, 0, 0)

            else:
                self.report({'WARNING'}, "Dust Particles Collection not found in the file")
                return {'CANCELLED'}

        except Exception as e:
            self.report({'ERROR'}, f"Failed to load Dust Particles Collection: {e}")
            return {'CANCELLED'}

        return {'FINISHED'}
        
class OBJECT_OT_DeleteDustParticles(Operator):
    bl_idname = "object.delete_dust_particles"
    bl_label = "Delete Dust Particles"
    bl_description = "Delete Dust Particles completely from the Scene"

    def execute(self, context):
        dust_particles_collection_name = "Dust Particles"

        context.scene.aged_fx.dust_particles_size_x = 0.0
        context.scene.aged_fx.dust_particles_size_y = 0.0
        context.scene.aged_fx.dust_particles_size_z = 0.0
        context.scene.aged_fx.dust_particles_wind_x = 0.0
        context.scene.aged_fx.dust_particles_wind_y = 0.0
        context.scene.aged_fx.dust_particles_wind_z = 0.0
        context.scene.aged_fx.dust_particles_amount = 0.0
        context.scene.aged_fx.dust_particles_scale_min = 0.0
        context.scene.aged_fx.dust_particles_scale_max = 0.0
        context.scene.aged_fx.dust_particles_speed = 0.0
        context.scene.aged_fx.dust_particles_rotation = 0.0
        context.scene.aged_fx.dust_particles_turbolence = 0.0
        context.scene.aged_fx.dust_particles_seed = 0
        context.scene.aged_fx.dust_particles_emission = 0.0

        if dust_particles_collection_name in bpy.data.collections:
            dust_particles_collection = bpy.data.collections[dust_particles_collection_name]

            for obj in dust_particles_collection.objects:
                bpy.data.objects.remove(obj, do_unlink=True)

            bpy.data.collections.remove(dust_particles_collection)

            for block in bpy.data.meshes:
                if block.users == 0:
                    bpy.data.meshes.remove(block)

        return {'FINISHED'}
        
class OBJECT_OT_AddMoths(bpy.types.Operator):
    bl_idname = "object.add_moths"
    bl_label = "Add Moths"
    bl_description = "Add Moths Emitter to the Scene"

    def execute(self, context):
        addon_prefs = context.preferences.addons[__package__].preferences
        filepath = addon_prefs.filepath

        if not context.selected_objects:
            self.report({'WARNING'}, "Select an object first.")
            return {'CANCELLED'}
        
        if len(context.selected_objects) > 1:
            self.report({'WARNING'}, "Only select one object.")
            return {'CANCELLED'}

        selected_object = context.selected_objects[0]

        moths_collection_name = "Moths AgedFX"
        moths_emitter_name = "Moths Emitter"

        moths_collection = bpy.data.collections.get(moths_collection_name)
        moths_emitter = bpy.data.objects.get(moths_emitter_name)

        if not moths_collection:
            with bpy.data.libraries.load(filepath, link=False) as (data_from, data_to):
                data_to.collections = [name for name in data_from.collections if name == moths_collection_name]

            moths_collection = bpy.data.collections.get(moths_collection_name)
            if moths_collection:
                context.scene.collection.children.link(moths_collection)
                moths_emitter = bpy.data.objects.get(moths_emitter_name)
                moths_emitter.location.x = selected_object.location.x
                moths_emitter.location.y = selected_object.location.y
                moths_emitter.location.z = 0

                view_layer = context.view_layer
                layer_collection = view_layer.layer_collection.children.get(moths_collection_name)
                if layer_collection:
                    for child in layer_collection.children:
                        if child.collection.name == "Moths Base":
                            child.exclude = True
                            break

                if moths_emitter:
                    settings = moths_emitter.particle_systems[0].settings
                    boids_part = settings.boids
                    state = boids_part.states[0]
                    rule = state.rules['Follow Leader']
                    rule.object = selected_object

                if "Collision" not in [mod.type for mod in selected_object.modifiers]:
                    selected_object.modifiers.new(name="Collision", type='COLLISION')

            else:
                self.report({'ERROR'}, "Moths AgedFX collection not found in the specified file.")
                return {'CANCELLED'}
        else:
            new_moths_emitter = moths_emitter.copy()
            new_moths_emitter.data = moths_emitter.data.copy()
            new_moths_emitter.animation_data_clear()
            moths_collection.objects.link(new_moths_emitter)

            new_moths_emitter.location.x = selected_object.location.x
            new_moths_emitter.location.y = selected_object.location.y
            new_moths_emitter.location.z = 0

            for ps in new_moths_emitter.particle_systems:
                new_particle_system = ps.settings.copy()
                ps.settings = new_particle_system

                settings = new_particle_system
                boids_part = settings.boids
                state = boids_part.states[0]
                rule = state.rules['Follow Leader']
                rule.object = selected_object

            if "Collision" not in [mod.type for mod in selected_object.modifiers]:
                selected_object.modifiers.new(name="Collision", type='COLLISION')

        return {'FINISHED'}

class OBJECT_OT_DeleteMoths(bpy.types.Operator):
    bl_idname = "object.delete_moths"
    bl_label = "Delete Moths"
    bl_description = "Delete selected Moths Emitters"

    def execute(self, context):
        if not context.selected_objects:
            self.report({'WARNING'}, "Select an object first.")
            return {'CANCELLED'}

        selected_object = context.selected_objects[0]
        moths_collection_name = "Moths AgedFX"
        moths_emitter_name = "Moths Emitter"

        moths_collection = bpy.data.collections.get(moths_collection_name)

        if not moths_collection:
            self.report({'ERROR'}, f"{moths_collection_name} collection not found.")
            return {'CANCELLED'}

        if selected_object.name not in moths_collection.objects:
            self.report({'WARNING'}, "Selected object is not part of the Moths AgedFX collection.")
            return {'CANCELLED'}

        for ps in selected_object.particle_systems:
            if ps.settings.name.startswith("Moths"):
                settings = selected_object.particle_systems[0].settings
                boids_part = settings.boids
                state = boids_part.states[0]
                rule = state.rules['Follow Leader']
                leader_object = rule.object
                
                if leader_object:
                    for mod in leader_object.modifiers:
                        if mod.type == 'COLLISION':
                            leader_object.modifiers.remove(mod)
                            break

        bpy.data.objects.remove(selected_object, do_unlink=True)

        remaining_emitters = [obj for obj in moths_collection.objects if obj.name.startswith(moths_emitter_name)]
        if not remaining_emitters:
            bpy.data.collections.remove(moths_collection)
            bpy.ops.outliner.orphans_purge(do_local_ids=True, do_linked_ids=True, do_recursive=True)

        return {'FINISHED'}
        
class OBJECT_OT_add_ants(bpy.types.Operator):
    bl_idname = "object.add_ants"
    bl_label = "Add Ants"

    def execute(self, context):
        prefs = context.preferences.addons[__package__].preferences
        file_path = prefs.filepath

        if "Ants AgedFX" not in bpy.data.collections:
            bpy.ops.wm.append(
                filepath=file_path,
                directory=os.path.join(file_path, "Collection"),
                filename="Ants AgedFX"
            )
            ants_collection = bpy.data.collections.get("Ants AgedFX")
            if ants_collection:
                ant_collection = ants_collection.children.get("Ant Base")
                if ant_collection:
                    ants_layer_collection = self.find_layer_collection(bpy.context.view_layer.layer_collection, "Ants AgedFX")
                    if ants_layer_collection:
                        ant_layer_collection = self.find_layer_collection(ants_layer_collection, "Ant Base")
                        if ant_layer_collection:
                            ant_layer_collection.exclude = True

        else:
            ants_obj = bpy.data.objects.get("Ants Emitter")
            if ants_obj:
                bpy.ops.object.select_all(action='DESELECT')
                ants_obj.select_set(True)
                bpy.ops.object.duplicate()
        return {'FINISHED'}

    def find_layer_collection(self, layer_collection, collection_name):
        if layer_collection.collection.name == collection_name:
            return layer_collection
        for child in layer_collection.children:
            found = self.find_layer_collection(child, collection_name)
            if found:
                return found
        return None

class OBJECT_OT_delete_ants(bpy.types.Operator):
    bl_idname = "object.delete_ants"
    bl_label = "Delete Ants"

    def execute(self, context):
        obj = context.active_object
        ants_collection = bpy.data.collections.get("Ants AgedFX")

        if obj and obj.name.startswith("Ants Emitter") and obj.name in ants_collection.objects:
            bpy.data.objects.remove(obj)
            if all(not o.name.startswith("Ants Emitter") for o in ants_collection.objects):
                bpy.data.collections.remove(ants_collection)
                bpy.ops.outliner.orphans_purge(do_local_ids=True, do_linked_ids=True, do_recursive=True)

        return {'FINISHED'}
        
class OBJECT_OT_add_shrinkwrap_plane_ants(bpy.types.Operator):
    bl_idname = "object.add_shrinkwrap_plane_ants"
    bl_label = "Add Shrinkwrap Plane"
    bl_options = {'REGISTER', 'UNDO'}
    
    def execute(self, context):
        ants_collection = bpy.data.collections.get("Ants AgedFX")
        
        if not ants_collection:
            self.report({'WARNING'}, "No ants have been added yet.")
            return {'CANCELLED'}
        
        if not context.selected_objects:
            self.report({'WARNING'}, "Please select a target object first.")
            return {'CANCELLED'}
        
        selected_obj = context.selected_objects[0]
        bpy.ops.mesh.primitive_plane_add()
        plane = context.object
        plane.name = "Shrinkwrap Plane"
        
        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.mesh.subdivide(number_cuts=100)
        bpy.ops.object.mode_set(mode='OBJECT')
        
        plane.location = selected_obj.location
        
        plane.scale[0] = selected_obj.dimensions[0] / 2
        plane.scale[1] = selected_obj.dimensions[1] / 2
        
        plane.scale[0] *= 1.2
        plane.scale[1] *= 1.2
        
        bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)
        
        shrinkwrap = plane.modifiers.new(name="Shrinkwrap", type='SHRINKWRAP')
        shrinkwrap.target = selected_obj
        
        for collection in plane.users_collection:
            collection.objects.unlink(plane)
        ants_collection.objects.link(plane)
        
        plane.display_type = 'WIRE'
        plane.hide_render = True
        
        return {'FINISHED'}
        
class OBJECT_OT_add_spiders(bpy.types.Operator):
    bl_idname = "object.add_spiders"
    bl_label = "Add Spiders"

    def execute(self, context):
        prefs = context.preferences.addons[__package__].preferences
        file_path = prefs.filepath

        if "Spiders AgedFX" not in bpy.data.collections:
            bpy.ops.wm.append(
                filepath=file_path,
                directory=os.path.join(file_path, "Collection"),
                filename="Spiders AgedFX"
            )
            spiders_collection = bpy.data.collections.get("Spiders AgedFX")
            if spiders_collection:
                spiders_layer_collection = self.find_layer_collection(bpy.context.view_layer.layer_collection, "Spiders AgedFX")
                if spiders_layer_collection:
                    for coll_name in ["Spider Controller", "Spider Base"]:
                        layer_to_exclude = self.find_layer_collection(spiders_layer_collection, coll_name)
                        if layer_to_exclude:
                            layer_to_exclude.exclude = True

        else:
            spiders_obj = bpy.data.objects.get("Spiders Emitter")
            if spiders_obj:
                bpy.ops.object.select_all(action='DESELECT')
                spiders_obj.select_set(True)
                bpy.ops.object.duplicate()
        return {'FINISHED'}

    def find_layer_collection(self, layer_collection, collection_name):
        if layer_collection.collection.name == collection_name:
            return layer_collection
        for child in layer_collection.children:
            found = self.find_layer_collection(child, collection_name)
            if found:
                return found
        return None

class OBJECT_OT_delete_spiders(bpy.types.Operator):
    bl_idname = "object.delete_spiders"
    bl_label = "Delete Spiders"

    def execute(self, context):
        obj = context.active_object
        spiders_collection = bpy.data.collections.get("Spiders AgedFX")

        if obj and obj.name.startswith("Spiders Emitter") and obj.name in spiders_collection.objects:
            bpy.data.objects.remove(obj)
            if all(not o.name.startswith("Spiders Emitter") for o in spiders_collection.objects):
                bpy.data.collections.remove(spiders_collection)
                bpy.ops.outliner.orphans_purge(do_local_ids=True, do_linked_ids=True, do_recursive=True)

        return {'FINISHED'}
        
class OBJECT_OT_add_shrinkwrap_plane_spiders(bpy.types.Operator):
    bl_idname = "object.add_shrinkwrap_plane_spiders"
    bl_label = "Add Shrinkwrap Plane"
    bl_options = {'REGISTER', 'UNDO'}
    
    def execute(self, context):
        ants_collection = bpy.data.collections.get("Spiders AgedFX")
        
        if not ants_collection:
            self.report({'WARNING'}, "No spiders have been added yet.")
            return {'CANCELLED'}
        
        if not context.selected_objects:
            self.report({'WARNING'}, "Please select a target object first.")
            return {'CANCELLED'}
        
        selected_obj = context.selected_objects[0]
        bpy.ops.mesh.primitive_plane_add()
        plane = context.object
        plane.name = "Shrinkwrap Plane"
        
        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.mesh.subdivide(number_cuts=100)
        bpy.ops.object.mode_set(mode='OBJECT')
        
        plane.location = selected_obj.location
        
        plane.scale[0] = selected_obj.dimensions[0] / 2
        plane.scale[1] = selected_obj.dimensions[1] / 2
        
        plane.scale[0] *= 1.2
        plane.scale[1] *= 1.2
        
        bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)
        
        shrinkwrap = plane.modifiers.new(name="Shrinkwrap", type='SHRINKWRAP')
        shrinkwrap.target = selected_obj
        
        for collection in plane.users_collection:
            collection.objects.unlink(plane)
        ants_collection.objects.link(plane)
        
        plane.display_type = 'WIRE'
        plane.hide_render = True
        
        return {'FINISHED'}
        
class OBJECT_OT_AddCollisionModifier(bpy.types.Operator):
    bl_idname = "object.add_collision_modifier"
    bl_label = "Add Collision Modifier"
    bl_description = "Add a Collision Modifier to selected objects"

    def execute(self, context):
        added_any = False
        
        for obj in context.selected_objects:
            if obj.type == 'MESH' and "Collision" not in [mod.type for mod in obj.modifiers]:
                obj.modifiers.new(name="Collision", type='COLLISION')
                added_any = True

        if added_any:
            self.report({'INFO'}, "Collision Modifier added to selected objects.")
        
        return {'FINISHED'}

class OBJECT_OT_RemoveCollisionModifier(bpy.types.Operator):
    bl_idname = "object.remove_collision_modifier"
    bl_label = "Remove Collision Modifier"
    bl_description = "Remove the Collision Modifier from selected objects"

    def execute(self, context):
        removed_any = False
        
        for obj in context.selected_objects:
            if obj.type == 'MESH':
                for mod in obj.modifiers:
                    if mod.type == 'COLLISION':
                        obj.modifiers.remove(mod)
                        removed_any = True
                        break

        if removed_any:
            self.report({'INFO'}, "Collision Modifier removed from selected objects.")

        return {'FINISHED'}
        
class OBJECT_OT_apply_transforms_selected(bpy.types.Operator):
    bl_idname = "object.apply_transforms_selected"
    bl_label = "Apply Transforms"

    def execute(self, context):
        obj = context.active_object
        if obj:
            bpy.ops.object.transform_apply(location=True, rotation=True, scale=True)
        return {'FINISHED'}
        
def toggle_volume_visibility(self, context):
    hide = self.hide_volume
    for obj in bpy.data.objects:
        if obj.name.startswith("Constant Light Ray Volume") or obj.name.startswith("Dissolving Light Ray Volume"):
            obj.hide_viewport = hide
        
class AgedFX_OT_AddConstantLightRay(Operator):
    bl_idname = "object.add_constant_light_ray"
    bl_label = "Add Constant Light Ray"

    def execute(self, context):
        prefs = context.preferences.addons[__package__].preferences
        file_path = prefs.filepath

        try:
            if "Constant Light Ray" not in bpy.data.collections:
                bpy.ops.wm.append(
                    filepath=file_path,
                    directory=os.path.join(file_path, "Collection"),
                    filename="Constant Light Ray"
                )
                if "Dissolving Light Ray" not in bpy.data.collections:
                    context.scene.aged_fx.light_ray_light = 80
                    context.scene.aged_fx.light_ray_volume = 0.2
                    context.scene.aged_fx.light_ray_color_stop_1 = (1.0, 1.0, 1.0, 1.0)
            else:
                self.duplicate_light_ray("Constant Light Ray")
        except Exception as e:
            self.report({'ERROR'}, f"Make sure you have set the right path in the addon preferences.")
        return {'FINISHED'}

    def duplicate_light_ray(self, collection_name):
        collection = bpy.data.collections.get(collection_name)
        if collection:
            for obj in collection.objects:
                if obj.name == collection_name:
                    new_obj = obj.copy()
                    new_obj.data = obj.data.copy()
                    collection.objects.link(new_obj)
                    for child in obj.children:
                        new_child = child.copy()
                        new_child.data = child.data.copy()
                        new_child.parent = new_obj
                        collection.objects.link(new_child)
        else:
            self.report({'ERROR'}, f"Collection '{collection_name}' not found")

class AgedFX_OT_AddDissolvingLightRay(Operator):
    bl_idname = "object.add_dissolving_light_ray"
    bl_label = "Add Dissolving Light Ray"

    def execute(self, context):
        prefs = context.preferences.addons[__package__].preferences
        file_path = prefs.filepath

        try:
            if "Dissolving Light Ray" not in bpy.data.collections:
                bpy.ops.wm.append(
                    filepath=file_path,
                    directory=os.path.join(file_path, "Collection"),
                    filename="Dissolving Light Ray"
                )
                if "Constant Light Ray" not in bpy.data.collections:
                    context.scene.aged_fx.light_ray_light = 80
                    context.scene.aged_fx.light_ray_volume = 0.2
                    context.scene.aged_fx.light_ray_color_stop_1 = (1.0, 1.0, 1.0, 1.0)
            else:
                self.duplicate_light_ray("Dissolving Light Ray")
        except Exception as e:
            self.report({'ERROR'}, f"Make sure you have set the right path in the addon preferences.")
        return {'FINISHED'}

    def duplicate_light_ray(self, collection_name):
        collection = bpy.data.collections.get(collection_name)
        if collection:
            for obj in collection.objects:
                if obj.name == collection_name:
                    new_obj = obj.copy()
                    new_obj.data = obj.data.copy()
                    collection.objects.link(new_obj)
                    for child in obj.children:
                        new_child = child.copy()
                        new_child.data = child.data.copy()
                        new_child.parent = new_obj
                        collection.objects.link(new_child)
        else:
            self.report({'ERROR'}, f"Collection '{collection_name}' not found")

class AgedFX_OT_DeleteLightRay(Operator):
    bl_idname = "object.delete_light_ray"
    bl_label = "Delete Light Ray"

    def execute(self, context):
        selected_obj = context.active_object
        if not selected_obj:
            self.report({'ERROR'}, "No object selected")
            return {'CANCELLED'}

        def remove_object_and_empty_collections(obj):
            if obj.users_collection:
                collection = obj.users_collection[0]
                bpy.data.objects.remove(obj, do_unlink=True)

                if collection.name in bpy.data.collections and len(collection.objects) == 0:
                    bpy.data.collections.remove(collection)
                    bpy.ops.outliner.orphans_purge(do_recursive=True)
                    return True
            return False

        if selected_obj.users_collection and selected_obj.users_collection[0].name in ["Constant Light Ray", "Dissolving Light Ray"]:
            collection = selected_obj.users_collection[0]
            parent = selected_obj.parent
            children = list(selected_obj.children)
            bpy.data.objects.remove(selected_obj, do_unlink=True)
            for child in children:
                bpy.data.objects.remove(child, do_unlink=True)

            if parent and len(parent.children) == 0:
                parent_deleted = remove_object_and_empty_collections(parent)
                if parent_deleted:
                    collection = None

            if collection and collection.name in bpy.data.collections and len(collection.objects) == 0:
                bpy.data.collections.remove(collection)
                bpy.ops.outliner.orphans_purge(do_recursive=True)

        if "Constant Light Ray" not in bpy.data.collections and "Dissolving Light Ray" not in bpy.data.collections:
            context.scene.aged_fx.light_ray_light = 80
            context.scene.aged_fx.light_ray_volume = 0.2
            context.scene.aged_fx.light_ray_color_stop_1 = (1.0, 1.0, 1.0, 1.0)

        return {'FINISHED'}
        
class OBJECT_OT_RemoveAllDriversConfirm(bpy.types.Operator):
    bl_idname = "object.remove_all_drivers_confirm"
    bl_label = "Remove All Drivers"
    bl_description = "Remove all drivers for dust, dirt, metal, and deformations properties on all objects in the scene"

    def execute(self, context):
        bpy.ops.object.remove_all_drivers('INVOKE_DEFAULT')
        return {'FINISHED'}

class OBJECT_OT_RemoveAllDrivers(bpy.types.Operator):
    bl_idname = "object.remove_all_drivers"
    bl_label = "Remove All Drivers"
    bl_description = "Remove all AgedFX related drivers and convert them into static numbers"

    def execute(self, context):
        properties_to_remove = {
            "Dust Amount", "Dust Scale", "Dust Top Ratio", "Dust Viewing Angle Unconsistency",
            "General Dirt Scale", "Edge Dirt Scale", "General Dirt", "Edge Dirt", "Edge Dirt Spread",
            "General Metal Scale", "Edge Metal Scale", "General Metal", "Edge Metal", "Edge Metal Spread",
            "Deformations Amount", "Deformations Scale", "Scratches", "Bumps"
        }

        aged_fx_drivers_to_remove = {
            "aged_fx.cobweb_translucency",
            "aged_fx.fog_density",
            "aged_fx.fog_anisotropy",
            "aged_fx.fog_x_scale",
            "aged_fx.fog_y_scale",
            "aged_fx.fog_z_scale",
            "aged_fx.ground_fog_density",
            "aged_fx.ground_fog_scale",
            "aged_fx.ground_fog_height",
            "aged_fx.ground_fog_position_x",
            "aged_fx.ground_fog_position_y",
            "aged_fx.ground_fog_speed_x",
            "aged_fx.ground_fog_speed_y",
            "aged_fx.ground_fog_turbolence",
            "aged_fx.dust_particles_size_x",
            "aged_fx.dust_particles_size_y",
            "aged_fx.dust_particles_size_z",
            "aged_fx.dust_particles_wind_x",
            "aged_fx.dust_particles_wind_y",
            "aged_fx.dust_particles_wind_z",
            "aged_fx.dust_particles_amount",
            "aged_fx.dust_particles_scale_min",
            "aged_fx.dust_particles_scale_max",
            "aged_fx.dust_particles_speed",
            "aged_fx.dust_particles_rotation",
            "aged_fx.dust_particles_turbolence",
            "aged_fx.dust_particles_seed",
            "aged_fx.dust_particles_emission",
            "aged_fx.light_ray_light",
            "aged_fx.light_ray_volume"
        }

        def remove_driver(node_tree, data_path):
            if node_tree.animation_data and node_tree.animation_data.drivers:
                for driver in node_tree.animation_data.drivers:
                    if driver.data_path == data_path:
                        node_tree.driver_remove(driver.data_path)

        def remove_driver_by_expression(node_tree, expressions):
            if node_tree.animation_data and node_tree.animation_data.drivers:
                drivers_to_remove = []
                for driver in node_tree.animation_data.drivers:
                    for var in driver.driver.variables:
                        if any(target.data_path in expressions for target in var.targets):
                            drivers_to_remove.append(driver)
                            break
                for driver in drivers_to_remove:
                    node_tree.driver_remove(driver.data_path)

        def remove_drivers_from_node_tree(node_tree):
            for node in node_tree.nodes:
                if node.type == 'VALUE':
                    if node.label in properties_to_remove:
                        data_path = f'nodes["{node.name}"].outputs[0].default_value'
                        remove_driver(node_tree, data_path)
            remove_driver_by_expression(node_tree, aged_fx_drivers_to_remove)

        def remove_drivers_from_material(material):
            if material.use_nodes:
                remove_drivers_from_node_tree(material.node_tree)

        def remove_drivers_from_geometry_nodes(modifier):
            if modifier.node_group:
                remove_drivers_from_node_tree(modifier.node_group)

        for obj in bpy.data.objects:
            if obj.type in {'MESH', 'CURVE'}:
                for slot in obj.material_slots:
                    material = slot.material
                    if material:
                        remove_drivers_from_material(material)
                if obj.type == 'MESH':
                    for modifier in obj.modifiers:
                        if modifier.type == 'NODES':
                            remove_drivers_from_geometry_nodes(modifier)

        self.report({'INFO'}, "All Drivers successfully deleted.")
        return {'FINISHED'}

    def invoke(self, context, event):
        return context.window_manager.invoke_confirm(self, event)
        
def find_input_name(node, input_name):
    for name in node.inputs.keys():
        if input_name in name:
            return name
    return None
                    
def material_uses_group(material, node_group_to_check) -> bool:
    for node in material.node_tree.nodes:
        if node.bl_idname == 'ShaderNodeGroup':
            if hasattr(node, 'node_tree') and node.node_tree.name == node_group_to_check.name:
                return True
    return False
                          
def get_materials_from_selected_objects():
    materials = set()
    for obj in bpy.context.selected_objects:
        if obj.type == 'MESH':
            for mat in obj.data.materials:
                materials.add(mat)
    return materials

def replace_glass_bsdf_in_materials(materials):
    for mat in materials:
        if mat.use_nodes:
            for node in mat.node_tree.nodes:
                if node.type == 'BSDF_GLASS':
                    replace_glass_with_principled(node, mat.node_tree)

def replace_glass_with_principled(glass_node, node_tree):
    principled_node = node_tree.nodes.new('ShaderNodeBsdfPrincipled')
    principled_node.location = glass_node.location

    input_mappings = {
        'Color': 'Base Color',
        'Roughness': 'Roughness',
        'IOR': 'IOR',
        'Normal': 'Normal'
    }

    for glass_input_name, principled_input_name in input_mappings.items():
        if glass_node.inputs[glass_input_name].is_linked:
            from_socket = glass_node.inputs[glass_input_name].links[0].from_socket
            node_tree.links.new(principled_node.inputs[principled_input_name], from_socket)
        else:
            principled_node.inputs[principled_input_name].default_value = glass_node.inputs[glass_input_name].default_value

    transmission_input_name = find_input_name(principled_node, 'Transmission')
    if transmission_input_name:
        principled_node.inputs[transmission_input_name].default_value = 1.0

    material_output = next((node for node in node_tree.nodes if node.type == 'OUTPUT_MATERIAL'), None)
    if material_output and glass_node.outputs['BSDF'].is_linked:
        node_tree.links.new(principled_node.outputs['BSDF'], material_output.inputs['Surface'])

    node_tree.nodes.remove(glass_node)
        
def is_object_in_excluded_collections(obj, excluded_collections):
    def collection_in_hierarchy(coll, excluded_collections):
        if coll.name in excluded_collections:
            return True
        for parent_coll in bpy.data.collections:
            if coll.name in [child.name for child in parent_coll.children]:
                if collection_in_hierarchy(parent_coll, excluded_collections):
                    return True
        return False

    for coll in obj.users_collection:
        if collection_in_hierarchy(coll, excluded_collections):
            return True
    return False

def make_dust(self, context):
    addon_prefs = context.preferences.addons[__package__].preferences
    filepath = addon_prefs.filepath

    my_node_group = bpy.data.node_groups.get('Dust', None)
    if not my_node_group:
        with bpy.data.libraries.load(filepath) as (data_from, data_to):
            data_to.node_groups = ['Dust']
        my_node_group = bpy.data.node_groups.get('Dust', None)
        
        if not my_node_group:
            self.report({'ERROR'}, "The Node Group 'Dust' could not be imported. Please make sure the path in the Addon Preferences is correct.")
            return {'CANCELLED'}

    excluded_collections = {"Moths AgedFX", "Ants AgedFX", "Spiders AgedFX"}
    objects = set(obj for obj in bpy.data.objects if obj.type == 'MESH' and not is_object_in_excluded_collections(obj, excluded_collections))
    materials_done = set()

    for obj in objects:
        for slot in obj.material_slots:
            if slot.material is None or slot.material.name in materials_done or not slot.material.use_nodes:
                continue

            if material_uses_group(slot.material, my_node_group):
                materials_done.add(slot.material.name)
                continue

            if add_node_group_to_material(slot.material, my_node_group, "aged_fx.dust_scale"):
                materials_done.add(slot.material.name)              
                if slot.material.use_nodes:
                    nodes = slot.material.node_tree.nodes
                    principled_bsdf = None
                    for node in nodes:
                        if node.type == 'BSDF_PRINCIPLED':
                            principled_bsdf = node
                            break

                    if principled_bsdf:
                        material = slot.material
                        material_output_node = next((node for node in material.node_tree.nodes if node.type == 'OUTPUT_MATERIAL'), None)
                        value_node = nodes.new(type="ShaderNodeValue")
                        value_node.label = "Dust Amount"
                        value_node.location = material_output_node.location + Vector((200, 0))

                        driver = value_node.outputs[0].driver_add("default_value").driver
                        driver.expression = "var"
                        var = driver.variables.new()
                        var.name = "var"
                        var.type = 'SINGLE_PROP'
                        var.targets[0].id_type = 'SCENE'
                        var.targets[0].id = context.scene
                        var.targets[0].data_path = "aged_fx.dust_amount"
                        
                        slot.material.node_tree.links.new(value_node.outputs[0], principled_bsdf.inputs['Sheen Weight'])

    return {'FINISHED'}
    
def make_dust_selected(self, context):
    addon_prefs = context.preferences.addons[__package__].preferences
    filepath = addon_prefs.filepath

    my_node_group = bpy.data.node_groups.get('Dust', None)
    if not my_node_group:
        with bpy.data.libraries.load(filepath) as (data_from, data_to):
            data_to.node_groups = ['Dust']
        my_node_group = bpy.data.node_groups.get('Dust', None)

        if not my_node_group:
            self.report({'ERROR'}, "The Node Group 'Dust' could not be imported. Please make sure the path in the Addon Preferences is correct.")
            return {'CANCELLED'}

    selected_objects = [obj for obj in context.selected_objects if obj.type == 'MESH']
    materials_done = set()

    for obj in selected_objects:
        for slot in obj.material_slots:
            if slot.material and slot.material.use_nodes and slot.material.name not in materials_done:
                if not any(node for node in slot.material.node_tree.nodes if node.type == 'GROUP' and node.node_tree == my_node_group):
                    if add_node_group_to_material(slot.material, my_node_group, "aged_fx.dust_scale"):
                        materials_done.add(slot.material.name)

                        nodes = slot.material.node_tree.nodes
                        principled_bsdf = None
                        for node in nodes:
                            if node.type == 'BSDF_PRINCIPLED':
                                principled_bsdf = node
                                break

                        if principled_bsdf:
                            material = slot.material
                            material_output_node = next((node for node in material.node_tree.nodes if node.type == 'OUTPUT_MATERIAL'), None)
                            value_node = nodes.new(type="ShaderNodeValue")
                            value_node.label = "Dust Amount"
                            value_node.location = material_output_node.location + Vector((200, 0))

                            driver = value_node.outputs[0].driver_add("default_value").driver
                            driver.expression = "var"
                            var = driver.variables.new()
                            var.name = "var"
                            var.type = 'SINGLE_PROP'
                            var.targets[0].id_type = 'SCENE'
                            var.targets[0].id = context.scene
                            var.targets[0].data_path = "aged_fx.dust_amount"
                            
                            slot.material.node_tree.links.new(value_node.outputs[0], principled_bsdf.inputs['Sheen Weight'])

    return {'FINISHED'}
    
def add_node_group_to_material(material, node_group_to_add, scale_property):
    pbsdf = None
    material_output = None
    for node in material.node_tree.nodes:
        if node.bl_idname == 'ShaderNodeBsdfPrincipled' and node.name not in ["Dirt_Principled_BSDF", "Metal_Principled_BSDF"]:
            pbsdf = node
        elif node.type == 'OUTPUT_MATERIAL':
            material_output = node

    if pbsdf is not None and material_output is not None:
        new_group = material.node_tree.nodes.new('ShaderNodeGroup')
        new_group.node_tree = node_group_to_add
        new_group.location = material_output.location + Vector([1400, 0])

        material.node_tree.links.new(pbsdf.inputs['Sheen Roughness'], new_group.outputs[0])
        pbsdf.inputs['Sheen Tint'].default_value = [0.5, 0.5, 0.5, 1.0]

        return True
    else:
        return False
    
def remove_node_group_from_material(material, node_group_to_remove):
    pbsdf = None
    for node in material.node_tree.nodes:
        if node.bl_idname == 'ShaderNodeBsdfPrincipled':
            pbsdf = node
            break

    if pbsdf is not None:
        for node in material.node_tree.nodes:
            if node.bl_idname == 'ShaderNodeGroup':
                if hasattr(node, 'node_tree') and node.node_tree.name == node_group_to_remove.name:
                    material.node_tree.nodes.remove(node)

        pbsdf.inputs['Sheen Weight'].driver_remove('default_value')

        pbsdf.inputs['Sheen Weight'].default_value = 0
        pbsdf.inputs['Sheen Tint'].default_value = [1.0, 1.0, 1.0, 1.0]

        return True
    else:
        return False

def add_dust_attributes_to_material(material):
    if not material.use_nodes:
        return

    nodes = material.node_tree.nodes
    links = material.node_tree.links

    principled_bsdf = next((node for node in nodes if node.type == 'BSDF_PRINCIPLED'), None)
    if not principled_bsdf:
        print(f"No Principled BSDF Node found in the material: {material.name}")
        return

    dust_group_node = next((node for node in nodes if node.type == 'GROUP' and node.node_tree.name == "Dust"), None)
    if not dust_group_node:
        print("Node Group 'Dust' not found in the material.")
        return

    material_output_node = next((node for node in material.node_tree.nodes if node.type == 'OUTPUT_MATERIAL'), None)

    value_nodes_info = [
        ("Dust Scale", "aged_fx.dust_scale", lambda: material_output_node.location + Vector((400, 0))),
        ("Dust Top Ratio", "aged_fx.dust_top", lambda: material_output_node.location + Vector((600, 0))),
        ("Dust Viewing Angle Unconsistency", "aged_fx.dust_side", lambda: material_output_node.location + Vector((800, 0))),
    ]

    for label, driver_path, position_func in value_nodes_info:
        if label not in nodes:
            value_node = nodes.new(type="ShaderNodeValue")
            value_node.name = value_node.label = label
            value_node.location = position_func()

            driver = value_node.outputs[0].driver_add("default_value").driver
            driver.expression = "var"
            var = driver.variables.new()
            var.name = "var"
            var.type = 'SINGLE_PROP'
            var.targets[0].id_type = 'SCENE'
            var.targets[0].id = bpy.context.scene
            var.targets[0].data_path = driver_path

            matched_input = next((input for input in dust_group_node.inputs if input.name == label), None)
            if matched_input:
                links.new(value_node.outputs[0], matched_input)
        else:
            print(f"Node '{label}' already exists in the material: {material.name}")
                
def remove_dust_attributes_from_material(material):
    if material.use_nodes:
        nodes = material.node_tree.nodes
        labels_to_remove = ["Dust Scale", "Dust Top Ratio", "Dust Viewing Angle Unconsistency", "Dust Top", "Dust Side Unvisibility"]

        for label in labels_to_remove:
            node_to_remove = next((node for node in nodes if node.label == label), None)
            if node_to_remove:
                nodes.remove(node_to_remove)
                print(f"Node with label '{label}' has been removed from the material {material.name}.")
            else:
                print(f"Node with label '{label}' not found in the material {material.name}.")
                
def find_original_principled_bsdf(material):
    for node in material.node_tree.nodes:
        if node.bl_idname == 'ShaderNodeBsdfPrincipled' and not node.name.startswith("Dirt_"):
            return node
    return None
                                
def add_dirt_attributes_to_material(material):
    if not material.use_nodes:
        return

    nodes = material.node_tree.nodes
    links = material.node_tree.links

    principled_bsdf = next((node for node in nodes if node.type == 'BSDF_PRINCIPLED'), None)
    if not principled_bsdf:
        print(f"No Principled BSDF Node found in the material: {material.name}")
        return

    dirt_group_node = next((node for node in nodes if node.type == 'GROUP' and node.node_tree.name == "Dirt"), None)
    if not dirt_group_node:
        print("Node Group 'Dirt' not found in the material.")
        return

    material_output_node = next((node for node in material.node_tree.nodes if node.type == 'OUTPUT_MATERIAL'), None)

    value_nodes_info = [
        ("General Dirt", "aged_fx.general_dirt", lambda: material_output_node.location + Vector((200, -150))),
        ("General Dirt Scale", "aged_fx.general_dirt_scale", lambda: material_output_node.location + Vector((400, -150))),
        ("Edge Dirt", "aged_fx.edge_dirt", lambda: material_output_node.location + Vector((600, -150))),
        ("Edge Dirt Scale", "aged_fx.edge_dirt_scale", lambda: material_output_node.location + Vector((800, -150))),
        ("Edge Dirt Spread", "aged_fx.edge_dirt_spread", lambda: material_output_node.location + Vector((1000, -150))),
    ]

    for label, driver_path, position_func in value_nodes_info:
        if label not in nodes:
            value_node = nodes.new(type="ShaderNodeValue")
            value_node.name = value_node.label = label
            value_node.location = position_func()

            driver = value_node.outputs[0].driver_add("default_value").driver
            driver.expression = "var"
            var = driver.variables.new()
            var.name = "var"
            var.type = 'SINGLE_PROP'
            var.targets[0].id_type = 'SCENE'
            var.targets[0].id = bpy.context.scene
            var.targets[0].data_path = driver_path

            matched_input = next((input for input in dirt_group_node.inputs if input.name == label), None)
            if matched_input:
                links.new(value_node.outputs[0], matched_input)
        else:
            print(f"Node '{label}' already exists in the material: {material.name}")
               
def remove_dirt_attributes_from_material(material):
    if material.use_nodes:
        nodes = material.node_tree.nodes
        labels_to_remove = ["Dirt Scale", "General Dirt Scale", "Edge Dirt Scale", "General Dirt", "Edge Dirt", "Edge Dirt Spread"]

        for label in labels_to_remove:
            node_to_remove = next((node for node in nodes if node.label == label), None)
            if node_to_remove:
                nodes.remove(node_to_remove)
                print(f"Node with label '{label}' has been removed from the material {material.name}.")
            else:
                print(f"Node with label '{label}' not found in the material {material.name}.")
                
def add_metal_attributes_to_material(material):
    if not material.use_nodes:
        return

    nodes = material.node_tree.nodes
    links = material.node_tree.links

    principled_bsdf = next((node for node in nodes if node.type == 'BSDF_PRINCIPLED'), None)
    if not principled_bsdf:
        print(f"No Principled BSDF Node found in the material: {material.name}")
        return

    metal_group_node = next((node for node in nodes if node.type == 'GROUP' and node.node_tree.name == "Metal"), None)
    if not metal_group_node:
        print("Node Group 'Metal' not found in the material.")
        return

    material_output_node = next((node for node in material.node_tree.nodes if node.type == 'OUTPUT_MATERIAL'), None)

    value_nodes_info = [
        ("General Metal", "aged_fx.general_metal", lambda: material_output_node.location + Vector((200, -300))),
        ("General Metal Scale", "aged_fx.general_metal_scale", lambda: material_output_node.location + Vector((400, -300))),
        ("Edge Metal", "aged_fx.edge_metal", lambda: material_output_node.location + Vector((600, -300))),
        ("Edge Metal Scale", "aged_fx.edge_metal_scale", lambda: material_output_node.location + Vector((800, -300))),
        ("Edge Metal Spread", "aged_fx.edge_metal_spread", lambda: material_output_node.location + Vector((1000, -300))),
    ]

    for label, driver_path, position_func in value_nodes_info:
        if label not in nodes:
            value_node = nodes.new(type="ShaderNodeValue")
            value_node.name = value_node.label = label
            value_node.location = position_func()

            driver = value_node.outputs[0].driver_add("default_value").driver
            driver.expression = "var"
            var = driver.variables.new()
            var.name = "var"
            var.type = 'SINGLE_PROP'
            var.targets[0].id_type = 'SCENE'
            var.targets[0].id = bpy.context.scene
            var.targets[0].data_path = driver_path

            matched_input = next((input for input in metal_group_node.inputs if input.name == label), None)
            if matched_input:
                links.new(value_node.outputs[0], matched_input)
        else:
            print(f"Node '{label}' already exists in the material: {material.name}")

def remove_metal_attributes_from_material(material):
    if material.use_nodes:
        nodes = material.node_tree.nodes
        labels_to_remove = ["General Metal Scale", "Edge Metal Scale", "General Metal", "Edge Metal", "Edge Metal Spread"]

        for label in labels_to_remove:
            node_to_remove = next((node for node in nodes if node.label == label), None)
            if node_to_remove:
                nodes.remove(node_to_remove)
                print(f"Node with label '{label}' has been removed from the material {material.name}.")
            else:
                print(f"Node with label '{label}' not found in the material {material.name}.")
                                   
def make_deformations(self, context):
    global deformations_group_imported
    if not deformations_group_imported:
        addon_prefs = context.preferences.addons[__package__].preferences
        filepath = addon_prefs.filepath
        with bpy.data.libraries.load(filepath) as (data_from, data_to):
            data_to.node_groups = ['Deformations']
        deformations_group_imported = True

    my_node_group = bpy.data.node_groups.get('Deformations', None)

    if not my_node_group:
        self.report({'ERROR'}, "The node group 'Deformations' could not be imported. Please ensure that the path in the Addon Preferences is correct.")
        return {'CANCELLED'}

    materials_done = set()

    for obj in bpy.data.objects:
        if obj.type == 'MESH':
            for slot in obj.material_slots:
                if slot.material is not None and slot.material.use_nodes:
                    if slot.material.name not in materials_done:
                        if add_deformations_to_material(slot.material, my_node_group):
                            materials_done.add(slot.material.name)

    return {'FINISHED'}
    
def make_deformations_selected(self, context):
    global deformations_group_imported
    if not deformations_group_imported:
        addon_prefs = context.preferences.addons[__package__].preferences
        filepath = addon_prefs.filepath
        with bpy.data.libraries.load(filepath) as (data_from, data_to):
            data_to.node_groups = ['Deformations']
        deformations_group_imported = True

    my_node_group = bpy.data.node_groups.get('Deformations', None)

    if not my_node_group:
        self.report({'ERROR'}, "The node group 'Deformations' could not be imported. Please ensure that the path in the Addon Preferences is correct.")
        return {'CANCELLED'}

    selected_objects = [obj for obj in context.selected_objects if obj.type == 'MESH']
    materials_done = set()

    for obj in selected_objects:
        for slot in obj.material_slots:
            if slot.material is not None and slot.material.use_nodes:
                if slot.material.name not in materials_done:
                    if add_deformations_to_material(slot.material, my_node_group):
                        materials_done.add(slot.material.name)

    return {'FINISHED'}

def delete_deformations(self, context):
    for obj in bpy.data.objects:
        if obj.type == 'MESH':
            for idx, slot in enumerate(obj.material_slots):
                if slot.material is not None and slot.material.use_nodes:
                    remove_deformations_from_material(slot.material)

    return {'FINISHED'}
    
def delete_deformations_selected(self, context):
    selected_objects = [obj for obj in context.selected_objects if obj.type == 'MESH']

    for obj in selected_objects:
        for idx, slot in enumerate(obj.material_slots):
            if slot.material is not None and slot.material.use_nodes:
                remove_deformations_from_material(slot.material)

    return {'FINISHED'}

def add_deformations_to_material(material, deformations_amount, materials_done):
    if material.name in materials_done:
        return

    pbsdf = find_original_principled_bsdf(material)
    if pbsdf is None:
        return

    material_output_node = next((node for node in material.node_tree.nodes if node.type == 'OUTPUT_MATERIAL'), None)
    if material_output_node is None:
        print(f"No Material Output Node found in the material {material.name}.")
        return

    mix_node = None
    normal_map_node = None
    deformations_node = None

    for node in material.node_tree.nodes:
        if node.bl_idname == 'ShaderNodeNormalMap':
            normal_map_node = node
        elif node.bl_idname == 'ShaderNodeMixRGB' and node.blend_type == 'ADD':
            mix_node = node
        elif node.bl_idname == 'ShaderNodeGroup' and node.node_tree and node.node_tree.name == 'Deformations':
            deformations_node = node

    if not mix_node or not deformations_node:
        if not normal_map_node:
            normal_map_node = material.node_tree.nodes.new('ShaderNodeNormalMap')
            normal_map_node.location = material_output_node.location + Vector([1800, -1050])
            normal_map_node.name = "Deformations_Normal_Map"

        if not mix_node:
            mix_node = material.node_tree.nodes.new('ShaderNodeMixRGB')
            mix_node.blend_type = 'ADD'
            mix_node.location = material_output_node.location + Vector([1600, -1050])
            material.node_tree.links.new(mix_node.inputs['Color1'], normal_map_node.outputs['Normal'])

            value_node = material.node_tree.nodes.new('ShaderNodeValue')
            value_node.label = "Deformations Amount"
            value_node.location = material_output_node.location + Vector([200, -450])

            driver = value_node.outputs[0].driver_add("default_value").driver
            driver.expression = "var"
            var = driver.variables.new()
            var.name = "var"
            var.type = 'SINGLE_PROP'
            var.targets[0].id_type = 'SCENE'
            var.targets[0].id = bpy.context.scene
            var.targets[0].data_path = "aged_fx.deformations_amount"

            material.node_tree.links.new(mix_node.inputs['Fac'], value_node.outputs[0])

        if not deformations_node:
            deformations_node = material.node_tree.nodes.new('ShaderNodeGroup')
            deformations_node.node_tree = bpy.data.node_groups['Deformations']
            deformations_node.location = material_output_node.location + Vector([1400, -1050])

        material.node_tree.links.new(mix_node.inputs['Color2'], deformations_node.outputs['Normal'])
        material.node_tree.links.new(pbsdf.inputs['Normal'], mix_node.outputs['Color'])

    materials_done.add(material.name)

def remove_deformations_from_material(material):
    for node in material.node_tree.nodes:
        if node.label == "Deformations Amount":
            material.node_tree.nodes.remove(node)
            break

    normal_map_node = None
    for node in material.node_tree.nodes:
        if node.bl_idname == 'ShaderNodeNormalMap' and node.name == 'Deformations_Normal_Map':
            normal_map_node = node
            break
    if normal_map_node:
        material.node_tree.nodes.remove(normal_map_node)

    nodes_to_remove = []
    for node in material.node_tree.nodes:
        if (node.bl_idname == 'ShaderNodeMixRGB' and node.blend_type == 'ADD') or \
           (node.bl_idname == 'ShaderNodeGroup' and node.node_tree and node.node_tree.name == 'Deformations'):
            nodes_to_remove.append(node)

    for node in nodes_to_remove:
        material.node_tree.nodes.remove(node)

    pbsdf = find_original_principled_bsdf(material)
    if pbsdf:
        for link in pbsdf.inputs['Normal'].links:
            material.node_tree.links.remove(link)

        for node in material.node_tree.nodes:
            if node.bl_idname == 'ShaderNodeNormalMap' and node.name != 'Deformations_Normal_Map':
                material.node_tree.links.new(pbsdf.inputs['Normal'], node.outputs['Normal'])
                break
                
def add_deformations_attributes_to_material(material):
    if not material.use_nodes:
        return

    nodes = material.node_tree.nodes
    links = material.node_tree.links

    principled_bsdf = next((node for node in nodes if node.type == 'BSDF_PRINCIPLED'), None)
    if not principled_bsdf:
        print(f"No Principled BSDF Node found in the material: {material.name}")
        return

    deformations_group_node = next((node for node in nodes if node.type == 'GROUP' and node.node_tree.name == "Deformations"), None)
    if not deformations_group_node:
        print("Node Group 'Deformations' not found in the material.")
        return
        
    material_output_node = next((node for node in material.node_tree.nodes if node.type == 'OUTPUT_MATERIAL'), None)

    value_nodes_info = [
        ("Deformations Scale", "aged_fx.deformations_scale", lambda: material_output_node.location + Vector((400, -450))),
        ("Scratches", "aged_fx.scratches", lambda: material_output_node.location + Vector((600, -450))),
        ("Bumps", "aged_fx.bumps", lambda: material_output_node.location + Vector((800, -450))),
    ]

    for label, driver_path, position_func in value_nodes_info:
        if label not in nodes:
            value_node = nodes.new(type="ShaderNodeValue")
            value_node.name = value_node.label = label
            value_node.location = position_func()

            driver = value_node.outputs[0].driver_add("default_value").driver
            driver.expression = "var"
            var = driver.variables.new()
            var.name = "var"
            var.type = 'SINGLE_PROP'
            var.targets[0].id_type = 'SCENE'
            var.targets[0].id = bpy.context.scene
            var.targets[0].data_path = driver_path

            matched_input = next((input for input in deformations_group_node.inputs if input.name == label), None)
            if matched_input:
                links.new(value_node.outputs[0], matched_input)
        else:
            print(f"Node '{label}' already exists in the material: {material.name}")
                
def remove_deformations_attributes_from_material(material):
    if material.use_nodes:
        nodes = material.node_tree.nodes
        labels_to_remove = ["Deformations Scale", "Scratches", "Bumps"]

        for label in labels_to_remove:
            node_to_remove = next((node for node in nodes if node.label == label), None)
            if node_to_remove:
                nodes.remove(node_to_remove)
                print(f"Node with label '{label}' has been removed from the material {material.name}.")
            else:
                print(f"Node with label '{label}' not found in the material {material.name}.")

def register():
    bpy.utils.register_class(AgedFXPreferences)
    bpy.utils.register_class(AgedFXProperties)
    bpy.utils.register_class(AgedFXPanel)
    bpy.utils.register_class(OBJECT_OT_MakeDustButton)
    bpy.utils.register_class(OBJECT_OT_DeleteDustButton)
    bpy.utils.register_class(OBJECT_OT_MakeDeformationsButton)
    bpy.utils.register_class(OBJECT_OT_DeleteDeformationsButton)
    bpy.utils.register_class(OBJECT_OT_MakeDirtButton)
    bpy.utils.register_class(OBJECT_OT_DeleteDirtButton)
    bpy.utils.register_class(OBJECT_OT_MakeDustSelectedButton)
    bpy.utils.register_class(OBJECT_OT_DeleteDustSelectedButton)
    bpy.utils.register_class(OBJECT_OT_MakeDirtSelectedButton)
    bpy.utils.register_class(OBJECT_OT_DeleteDirtSelectedButton)
    bpy.utils.register_class(OBJECT_OT_MakeDeformationsSelectedButton)
    bpy.utils.register_class(OBJECT_OT_DeleteDeformationsSelectedButton)
    bpy.utils.register_class(OBJECT_OT_ReplaceGlassAll)
    bpy.utils.register_class(OBJECT_OT_ReplaceGlassSelected)
    bpy.utils.register_class(OBJECT_OT_ApplyScale)
    bpy.utils.register_class(OBJECT_OT_AddCobweb)
    bpy.utils.register_class(OBJECT_OT_DeleteCobweb)
    bpy.utils.register_class(OBJECT_OT_AddFog)
    bpy.utils.register_class(OBJECT_OT_DeleteFog)
    bpy.utils.register_class(OBJECT_OT_AddDustParticles)
    bpy.utils.register_class(OBJECT_OT_DeleteDustParticles)
    bpy.utils.register_class(OBJECT_OT_MakeSingleUserData)
    bpy.utils.register_class(OBJECT_OT_MakeUniqueDust)
    bpy.utils.register_class(OBJECT_OT_MakeUniqueDirt)
    bpy.utils.register_class(OBJECT_OT_MakeUniqueDeformations)
    bpy.utils.register_class(OBJECT_OT_AddMossGenerator)
    bpy.utils.register_class(OBJECT_OT_RemoveMossGenerator)
    bpy.utils.register_class(OBJECT_OT_toggle_apply_scale)
    bpy.utils.register_class(OBJECT_OT_ConvertSelectedToMesh)
    bpy.utils.register_class(OBJECT_OT_MakeMetalButton)
    bpy.utils.register_class(OBJECT_OT_DeleteMetalButton)
    bpy.utils.register_class(OBJECT_OT_MakeMetalSelectedButton)
    bpy.utils.register_class(OBJECT_OT_DeleteMetalSelectedButton)
    bpy.utils.register_class(OBJECT_OT_MakeUniqueMetal)
    bpy.utils.register_class(OBJECT_OT_MakeSelectedRustyButton)
    bpy.utils.register_class(OBJECT_OT_UndoRustyButton)
    bpy.utils.register_class(OBJECT_OT_AddMoths)
    bpy.utils.register_class(OBJECT_OT_DeleteMoths)
    bpy.utils.register_class(OBJECT_OT_AddCollisionModifier)
    bpy.utils.register_class(OBJECT_OT_RemoveCollisionModifier)
    bpy.utils.register_class(OBJECT_OT_add_ants)
    bpy.utils.register_class(OBJECT_OT_delete_ants)
    bpy.utils.register_class(OBJECT_OT_add_spiders)
    bpy.utils.register_class(OBJECT_OT_delete_spiders)
    bpy.utils.register_class(OBJECT_OT_apply_transforms_selected)
    bpy.utils.register_class(AgedFX_OT_AddConstantLightRay)
    bpy.utils.register_class(AgedFX_OT_AddDissolvingLightRay)
    bpy.utils.register_class(AgedFX_OT_DeleteLightRay)
    bpy.utils.register_class(OBJECT_OT_RemoveAllDrivers)
    bpy.utils.register_class(OBJECT_OT_add_shrinkwrap_plane_ants)
    bpy.utils.register_class(OBJECT_OT_add_shrinkwrap_plane_spiders)
    bpy.types.Scene.aged_fx = PointerProperty(type=AgedFXProperties)
    bpy.app.translations.register(__package__, GetTranslationDict())
    print(f"----  {Path(__file__).parent.name} 插件由xz-blender购买分享!感谢使用!  ----")

def unregister():
    del bpy.types.Scene.aged_fx

    bpy.utils.unregister_class(AgedFXPreferences)
    bpy.utils.unregister_class(AgedFXProperties)
    bpy.utils.unregister_class(AgedFXPanel)
    bpy.utils.unregister_class(OBJECT_OT_MakeDustButton)
    bpy.utils.unregister_class(OBJECT_OT_DeleteDustButton)
    bpy.utils.unregister_class(OBJECT_OT_MakeDeformationsButton)
    bpy.utils.unregister_class(OBJECT_OT_DeleteDeformationsButton)
    bpy.utils.unregister_class(OBJECT_OT_MakeDirtButton)
    bpy.utils.unregister_class(OBJECT_OT_DeleteDirtButton)
    bpy.utils.unregister_class(OBJECT_OT_MakeDustSelectedButton)
    bpy.utils.unregister_class(OBJECT_OT_DeleteDustSelectedButton)
    bpy.utils.unregister_class(OBJECT_OT_MakeDirtSelectedButton)
    bpy.utils.unregister_class(OBJECT_OT_DeleteDirtSelectedButton)
    bpy.utils.unregister_class(OBJECT_OT_MakeDeformationsSelectedButton)
    bpy.utils.unregister_class(OBJECT_OT_DeleteDeformationsSelectedButton)
    bpy.utils.unregister_class(OBJECT_OT_ReplaceGlassAll)
    bpy.utils.unregister_class(OBJECT_OT_ReplaceGlassSelected)
    bpy.utils.unregister_class(OBJECT_OT_ApplyScale)
    bpy.utils.unregister_class(OBJECT_OT_AddCobweb)
    bpy.utils.unregister_class(OBJECT_OT_DeleteCobweb)
    bpy.utils.unregister_class(OBJECT_OT_AddFog)
    bpy.utils.unregister_class(OBJECT_OT_DeleteFog)
    bpy.utils.unregister_class(OBJECT_OT_AddDustParticles)
    bpy.utils.unregister_class(OBJECT_OT_DeleteDustParticles)
    bpy.utils.unregister_class(OBJECT_OT_MakeSingleUserData)
    bpy.utils.unregister_class(OBJECT_OT_MakeUniqueDust)
    bpy.utils.unregister_class(OBJECT_OT_MakeUniqueDirt)
    bpy.utils.unregister_class(OBJECT_OT_MakeUniqueDeformations)
    bpy.utils.unregister_class(OBJECT_OT_AddMossGenerator)
    bpy.utils.unregister_class(OBJECT_OT_RemoveMossGenerator)
    bpy.utils.unregister_class(OBJECT_OT_toggle_apply_scale)
    bpy.utils.unregister_class(OBJECT_OT_ConvertSelectedToMesh)
    bpy.utils.unregister_class(OBJECT_OT_MakeMetalButton)
    bpy.utils.unregister_class(OBJECT_OT_DeleteMetalButton)
    bpy.utils.unregister_class(OBJECT_OT_MakeMetalSelectedButton)
    bpy.utils.unregister_class(OBJECT_OT_DeleteMetalSelectedButton)
    bpy.utils.unregister_class(OBJECT_OT_MakeUniqueMetal)
    bpy.utils.unregister_class(OBJECT_OT_MakeSelectedRustyButton)
    bpy.utils.unregister_class(OBJECT_OT_UndoRustyButton)
    bpy.utils.unregister_class(OBJECT_OT_AddMoths)
    bpy.utils.unregister_class(OBJECT_OT_DeleteMoths)
    bpy.utils.unregister_class(OBJECT_OT_AddCollisionModifier)
    bpy.utils.unregister_class(OBJECT_OT_RemoveCollisionModifier)
    bpy.utils.unregister_class(OBJECT_OT_add_ants)
    bpy.utils.unregister_class(OBJECT_OT_delete_ants)
    bpy.utils.unregister_class(OBJECT_OT_add_spiders)
    bpy.utils.unregister_class(OBJECT_OT_delete_spiders)
    bpy.utils.unregister_class(OBJECT_OT_apply_transforms_selected)
    bpy.utils.unregister_class(AgedFX_OT_AddConstantLightRay)
    bpy.utils.unregister_class(AgedFX_OT_AddDissolvingLightRay)
    bpy.utils.unregister_class(AgedFX_OT_DeleteLightRay)
    bpy.utils.unregister_class(OBJECT_OT_RemoveAllDrivers)
    bpy.utils.unregister_class(OBJECT_OT_add_shrinkwrap_plane_ants)
    bpy.utils.unregister_class(OBJECT_OT_add_shrinkwrap_plane_spiders)
    bpy.app.translations.unregister(__package__)

if __name__ == "__main__":
    register()